# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides models for doing PSF/PRF-fitting photometry.
"""

import copy
import itertools
import warnings

from astropy.modeling import Fittable2DModel, Parameter
from astropy.nddata import NDData
from astropy.utils.exceptions import AstropyWarning
import numpy as np

from ..aperture import CircularAperture

__all__ = ['NonNormalizable', 'FittableImageModel', 'EPSFModel',
           'GriddedPSFModel', 'IntegratedGaussianPRF', 'PRFAdapter']


class NonNormalizable(AstropyWarning):
    """
    Used to indicate that a :py:class:`FittableImageModel` model is
    non-normalizable.
    """


class FittableImageModel(Fittable2DModel):
    r"""
    A fittable 2D model of an image allowing for image intensity scaling
    and image translations.

    This class takes 2D image data and computes the values of the model
    at arbitrary locations (including at intra-pixel, fractional
    positions) within this image using spline interpolation provided by
    :py:class:`~scipy.interpolate.RectBivariateSpline`.

    The fittable model provided by this class has three model
    parameters: an image intensity scaling factor (``flux``) which is
    applied to (normalized) image, and two positional parameters
    (``x_0`` and ``y_0``) indicating the location of a feature in the
    coordinate grid on which the model is to be evaluated.

    If this class is initialized with ``flux`` (intensity scaling
    factor) set to `None`, then ``flux`` is going to be estimated as
    ``sum(data)``.

    Parameters
    ----------
    data : numpy.ndarray
        Array containing 2D image.

    origin : tuple, None, optional
        A reference point in the input image ``data`` array. When origin
        is `None`, origin will be set at the middle of the image array.

        If ``origin`` represents the location of a feature (e.g., the
        position of an intensity peak) in the input ``data``, then model
        parameters ``x_0`` and ``y_0`` show the location of this peak in
        an another target image to which this model was fitted.
        Fundamentally, it is the coordinate in the model's image data
        that should map to coordinate (``x_0``, ``y_0``) of the output
        coordinate system on which the model is evaluated.

        Alternatively, when ``origin`` is set to ``(0,0)``, then model
        parameters ``x_0`` and ``y_0`` are shifts by which model's image
        should be translated in order to match a target image.

    normalize : bool, optional
        Indicates whether or not the model should be build on normalized
        input image data. If true, then the normalization constant (*N*)
        is computed so that

        .. math::
            N \cdot C \cdot \sum\limits_{i,j} D_{i,j} = 1,

        where *N* is the normalization constant, *C* is correction
        factor given by the parameter ``normalization_correction``, and
        :math:`D_{i,j}` are the elements of the input image ``data``
        array.

    normalization_correction : float, optional
        A strictly positive number that represents correction that needs
        to be applied to model's data normalization (see *C* in the
        equation in the comments to ``normalize`` for more details).  A
        possible application for this parameter is to account for
        aperture correction. Assuming model's data represent a PSF to be
        fitted to some target star, we set ``normalization_correction``
        to the aperture correction that needs to be applied to the
        model. That is, ``normalization_correction`` in this case should
        be set to the ratio between the total flux of the PSF (including
        flux outside model's data) to the flux of model's data.  Then,
        best fitted value of the ``flux`` model parameter will represent
        an aperture-corrected flux of the target star.  In the case of
        aperture correction, ``normalization_correction`` should be a
        value larger than one, as the total flux, including regions
        outside of the aperture, should be larger than the flux inside
        the aperture, and thus the correction is applied as an inversely
        multiplied factor.

    fill_value : float, optional
        The value to be returned by the `evaluate` or
        ``astropy.modeling.Model.__call__`` methods when evaluation is
        performed outside the definition domain of the model.

    kwargs : dict, optional
        Additional optional keyword arguments to be passed directly to
        the `compute_interpolator` method.  See `compute_interpolator`
        for more details.

    oversampling : int or tuple of two int, optional
        The oversampling factor(s) of the model in the ``x`` and ``y``
        directions.  If ``oversampling`` is a scalar it will be treated
        as being the same in both x and y; otherwise a tuple of two
        floats will be treated as ``(x_oversamp, y_oversamp)``.
    """

    flux = Parameter(description='Intensity scaling factor for image data.',
                     default=1.0)
    x_0 = Parameter(description='X-position of a feature in the image in '
                    'the output coordinate grid on which the model is '
                    'evaluated.', default=0.0)
    y_0 = Parameter(description='Y-position of a feature in the image in '
                    'the output coordinate grid on which the model is '
                    'evaluated.', default=0.0)

    def __init__(self, data, flux=flux.default, x_0=x_0.default,
                 y_0=y_0.default, normalize=False,
                 normalization_correction=1.0, origin=None, oversampling=1,
                 fill_value=0.0, **kwargs):

        self._fill_value = fill_value
        self._img_norm = None
        self._normalization_status = 0 if normalize else 2
        self._store_interpolator_kwargs(**kwargs)
        self._set_oversampling(oversampling)

        if normalization_correction <= 0:
            raise ValueError("'normalization_correction' must be strictly "
                             "positive.")
        self._normalization_correction = normalization_correction

        self._data = np.array(data, copy=True, dtype=float)

        if not np.all(np.isfinite(self._data)):
            raise ValueError("All elements of input 'data' must be finite.")

        # set input image related parameters:
        self._ny, self._nx = self._data.shape
        self._shape = self._data.shape
        if self._data.size < 1:
            raise ValueError("Image data array cannot be zero-sized.")

        # set the origin of the coordinate system in image's pixel grid:
        self.origin = origin

        flux = self._initial_norm(flux, normalize)

        super().__init__(flux, x_0, y_0)

        # initialize interpolator:
        self.compute_interpolator(**kwargs)

    def _initial_norm(self, flux, normalize):

        if flux is None:
            if self._img_norm is None:
                self._img_norm = self._compute_raw_image_norm()
            flux = self._img_norm

        self._compute_normalization(normalize)

        return flux

    def _compute_raw_image_norm(self):
        """
        Helper function that computes the uncorrected inverse
        normalization factor of input image data. This quantity is
        computed as the *sum of all pixel values*.

        .. note::
            This function is intended to be overridden in a subclass if
            one desires to change the way the normalization factor is
            computed.
        """
        return np.sum(self._data, dtype=float)

    def _compute_normalization(self, normalize):
        r"""
        Helper function that computes (corrected) normalization factor
        of the original image data. This quantity is computed as the
        inverse "raw image norm" (or total "flux" of model's image)
        corrected by the ``normalization_correction``:

        .. math::
            N = 1/(\Phi * C),

        where :math:`\Phi` is the "total flux" of model's image as
        computed by `_compute_raw_image_norm` and *C* is the
        normalization correction factor. :math:`\Phi` is computed only
        once if it has not been previously computed. Otherwise, the
        existing (stored) value of :math:`\Phi` is not modified as
        :py:class:`FittableImageModel` does not allow image data to be
        modified after the object is created.

        .. note::
            Normally, this function should not be called by the
            end-user. It is intended to be overridden in a subclass if
            one desires to change the way the normalization factor is
            computed.
        """
        self._normalization_constant = 1.0 / self._normalization_correction

        if normalize:
            # compute normalization constant so that
            # N*C*sum(data) = 1:
            if self._img_norm is None:
                self._img_norm = self._compute_raw_image_norm()

            if self._img_norm != 0.0 and np.isfinite(self._img_norm):
                self._normalization_constant /= self._img_norm
                self._normalization_status = 0

            else:
                self._normalization_constant = 1.0
                self._normalization_status = 1
                warnings.warn("Overflow encountered while computing "
                              "normalization constant. Normalization "
                              "constant will be set to 1.", NonNormalizable)

        else:
            self._normalization_status = 2

    @property
    def oversampling(self):
        """
        The factor by which the stored image is oversampled.

        An input to this model is multiplied by this factor to yield the
        index into the stored image.
        """
        return self._oversampling

    def _set_oversampling(self, value):
        value = np.atleast_1d(value)

        if np.any(~np.isfinite(value)):
            raise ValueError('Oversampling factor must be a finite value')

        if np.any(value <= 0):
            raise ValueError('Oversampling factor must be greater than 0')

        if len(value) == 1:
            value = np.repeat(value, 2)
        if len(value) != 2:
            raise ValueError('Oversampling factor must have 1 or 2 '
                             'elements')
        if value.ndim != 1:
            raise ValueError('Oversampling factor must be 1D')

        value_int = value.astype(int)
        if np.any(value_int != value):  # e.g., 2.0 is OK, 2.1 is not
            raise ValueError('Oversampling elements must be integers')

        self._oversampling = value

    @property
    def data(self):
        """Get original image data."""
        return self._data

    @property
    def normalized_data(self):
        """Get normalized and/or intensity-corrected image data."""
        return self._normalization_constant * self._data

    @property
    def normalization_constant(self):
        """Get normalization constant."""
        return self._normalization_constant

    @property
    def normalization_status(self):
        """
        Get normalization status. Possible status values are:

        - 0: **Performed**. Model has been successfully normalized at
             user's request.
        - 1: **Failed**. Attempt to normalize has failed.
        - 2: **NotRequested**. User did not request model to be normalized.
        """
        return self._normalization_status

    @property
    def normalization_correction(self):
        """
        Set/Get flux correction factor.

        .. note::
            When setting correction factor, model's flux will be
            adjusted accordingly such that if this model was a good fit
            to some target image before, then it will remain a good fit
            after correction factor change.
        """
        return self._normalization_correction

    @normalization_correction.setter
    def normalization_correction(self, normalization_correction):
        old_cf = self._normalization_correction
        self._normalization_correction = normalization_correction
        self._compute_normalization(normalize=self._normalization_status != 2)

        # adjust model's flux so that if this model was a good fit to
        # some target image, then it will remain a good fit after
        # correction factor change:
        self.flux *= normalization_correction / old_cf

    @property
    def shape(self):
        """
        A tuple of dimensions of the data array in numpy style (ny, nx).
        """
        return self._shape

    @property
    def nx(self):
        """Number of columns in the data array."""
        return self._nx

    @property
    def ny(self):
        """Number of rows in the data array."""
        return self._ny

    @property
    def origin(self):
        """
        A tuple of ``x`` and ``y`` coordinates of the origin of the
        coordinate system in terms of pixels of model's image.

        When setting the coordinate system origin, a tuple of two `int`
        or `float` may be used. If origin is set to `None`, the origin
        of the coordinate system will be set to the middle of the data
        array (``(npix-1)/2.0``).

        .. warning::
            Modifying ``origin`` will not adjust (modify) model's
            parameters ``x_0`` and ``y_0``.
        """
        return (self._x_origin, self._y_origin)

    @origin.setter
    def origin(self, origin):
        if origin is None:
            self._x_origin = (self._nx - 1) / 2.0
            self._y_origin = (self._ny - 1) / 2.0
        elif hasattr(origin, '__iter__') and len(origin) == 2:
            self._x_origin, self._y_origin = origin
        else:
            raise TypeError("Parameter 'origin' must be either None or an "
                            "iterable with two elements.")

    @property
    def x_origin(self):
        """X-coordinate of the origin of the coordinate system."""
        return self._x_origin

    @property
    def y_origin(self):
        """Y-coordinate of the origin of the coordinate system."""
        return self._y_origin

    @property
    def fill_value(self):
        """
        Fill value to be returned for coordinates outside of the domain
        of definition of the interpolator. If ``fill_value`` is `None`,
        then values outside of the domain of definition are the ones
        returned by the interpolator.
        """
        return self._fill_value

    @fill_value.setter
    def fill_value(self, fill_value):
        self._fill_value = fill_value

    def _store_interpolator_kwargs(self, **kwargs):
        """
        This function should be called in a subclass whenever model's
        interpolator is (re-)computed.
        """
        self._interpolator_kwargs = copy.deepcopy(kwargs)

    @property
    def interpolator_kwargs(self):
        """
        Get current interpolator's arguments used when interpolator was
        created.
        """
        return self._interpolator_kwargs

    def compute_interpolator(self, **kwargs):
        """
        Compute/define the interpolating spline.

        This function can be overridden in a subclass to define custom
        interpolators.

        Parameters
        ----------
        **kwargs : dict, optional
            Additional optional keyword arguments:

            - **degree** : int, tuple, optional
                Degree of the interpolating spline. A tuple can be used
                to provide different degrees for the X- and Y-axes.
                Default value is degree=3.

            - **s** : float, optional
                Non-negative smoothing factor. Default value s=0
                corresponds to interpolation.  See
                :py:class:`~scipy.interpolate.RectBivariateSpline` for
                more details.

        Notes
        -----
        * When subclassing :py:class:`FittableImageModel` for the
          purpose of overriding :py:func:`compute_interpolator`, the
          :py:func:`evaluate` may need to overridden as well depending
          on the behavior of the new interpolator. In addition, for
          improved future compatibility, make sure that the overriding
          method stores keyword arguments ``kwargs`` by calling
          ``_store_interpolator_kwargs`` method.

        * Use caution when modifying interpolator's degree or smoothness
          in a computationally intensive part of the code as it may
          decrease code performance due to the need to recompute
          interpolator.
        """
        from scipy.interpolate import RectBivariateSpline

        if 'degree' in kwargs:
            degree = kwargs['degree']
            if hasattr(degree, '__iter__') and len(degree) == 2:
                degx = int(degree[0])
                degy = int(degree[1])
            else:
                degx = int(degree)
                degy = int(degree)
            if degx < 0 or degy < 0:
                raise ValueError("Interpolator degree must be a non-negative "
                                 "integer")
        else:
            degx = 3
            degy = 3

        smoothness = kwargs.get('s', 0)

        x = np.arange(self._nx, dtype=float)
        y = np.arange(self._ny, dtype=float)
        self.interpolator = RectBivariateSpline(
            x, y, self._data.T, kx=degx, ky=degy, s=smoothness
        )

        self._store_interpolator_kwargs(**kwargs)

    def evaluate(self, x, y, flux, x_0, y_0, use_oversampling=True):
        """
        Evaluate the model on some input variables and provided model
        parameters.

        Parameters
        ----------
        use_oversampling : bool, optional
            Whether to use the oversampling factor to calculate the
            model pixel indices. The default is `True`, which means the
            input indices will be multiplied by this factor.
        """
        if use_oversampling:
            xi = self._oversampling[0] * (np.asarray(x) - x_0)
            yi = self._oversampling[1] * (np.asarray(y) - y_0)
        else:
            xi = np.asarray(x) - x_0
            yi = np.asarray(y) - y_0

        xi = xi.astype(float)
        yi = yi.astype(float)
        xi += self._x_origin
        yi += self._y_origin

        f = flux * self._normalization_constant
        evaluated_model = f * self.interpolator.ev(xi, yi)

        if self._fill_value is not None:
            # find indices of pixels that are outside the input pixel grid and
            # set these pixels to the 'fill_value':
            invalid = (((xi < 0) | (xi > self._nx - 1)) |
                       ((yi < 0) | (yi > self._ny - 1)))
            evaluated_model[invalid] = self._fill_value

        return evaluated_model


class EPSFModel(FittableImageModel):
    """
    A class that models an effective PSF (ePSF).

    The EPSFModel is normalized such that the sum of the PSF over the
    (undersampled) pixels within the the input ``norm_radius`` is 1.0.
    This means that when the EPSF is fit to stars, the resulting flux
    corresponds to aperture photometry within a circular aperture of
    radius ``norm_radius``.

    While this class is a subclass of `FittableImageModel`, it is very
    similar.  The primary differences/motivation are a few additional
    parameters necessary specifically for ePSFs.

    Parameters
    ----------
    oversampling : int or tuple of two int, optional
        The oversampling factor(s) of the model in the ``x`` and ``y``
        directions.  If ``oversampling`` is a scalar it will be treated
        as being the same in both x and y; otherwise a tuple of two
        floats will be treated as ``(x_oversamp, y_oversamp)``.

    norm_radius : float, optional
        The radius inside which the ePSF is normalized by the sum over
        undersampled integer pixel values inside a circular aperture.

    shift_val : float, optional
        The fractional undersampled pixel amount (equivalent to an
        integer oversampled pixel value) at which to evaluate the
        asymmetric ePSF centroid corrections.
    """

    def __init__(self, data, flux=1.0, x_0=0.0, y_0=0.0, normalize=True,
                 normalization_correction=1.0, origin=None, oversampling=1,
                 fill_value=0.0, norm_radius=5.5, shift_val=0.5, **kwargs):

        self._norm_radius = norm_radius
        self._shift_val = shift_val

        super().__init__(data=data, flux=flux, x_0=x_0, y_0=y_0,
                         normalize=normalize,
                         normalization_correction=normalization_correction,
                         origin=origin, oversampling=oversampling,
                         fill_value=fill_value, **kwargs)

    def _initial_norm(self, flux, normalize):
        if flux is None:
            if self._img_norm is None:
                self._img_norm = self._compute_raw_image_norm()
            flux = self._img_norm

        if normalize:
            self._compute_normalization()
        else:
            self._img_norm = self._compute_raw_image_norm()

        return flux

    def _compute_raw_image_norm(self):
        """
        Compute the normalization of input image data as the flux
        within a given radius.
        """
        xypos = (self._nx / 2., self._ny / 2.)
        # TODO: generalize "radius" (ellipse?) is oversampling is
        # different along x/y axes
        radius = self._norm_radius * self.oversampling[0]
        aper = CircularAperture(xypos, r=radius)
        flux, _ = aper.do_photometry(self._data, method='exact')
        return flux[0] / np.product(self.oversampling)

    def _compute_normalization(self):
        """
        Helper function that computes (corrected) normalization factor
        of the original image data. For the ePSF this is defined as the
        sum over the inner N (default=5.5) pixels of the non-oversampled
        image. Will re-normalize the data to the value calculated.
        """
        if self._img_norm is None:
            if np.sum(self._data) == 0:
                self._img_norm = 1
            else:
                self._img_norm = self._compute_raw_image_norm()

        if self._img_norm != 0.0 and np.isfinite(self._img_norm):
            self._data /= (self._img_norm * self._normalization_correction)
            self._normalization_status = 0
        else:
            self._normalization_status = 1
            self._img_norm = 1
            warnings.warn("Overflow encountered while computing "
                          "normalization constant. Normalization "
                          "constant will be set to 1.", NonNormalizable)

    def normalized_data(self):
        """
        Overloaded dummy function that also returns self._data, as the
        normalization occurs within _compute_normalization in EPSFModel,
        and as such self._data will sum, accounting for
        under/oversampled pixels, to 1/self._normalization_correction.
        """
        return self._data

    @FittableImageModel.origin.setter
    def origin(self, origin):
        if origin is None:
            self._x_origin = (self._nx - 1) / 2.0 / self.oversampling[0]
            self._y_origin = (self._ny - 1) / 2.0 / self.oversampling[1]
        elif (hasattr(origin, '__iter__') and len(origin) == 2):
            self._x_origin, self._y_origin = origin
        else:
            raise TypeError("Parameter 'origin' must be either None or an "
                            "iterable with two elements.")

    def compute_interpolator(self, **kwargs):
        """
        Compute/define the interpolating spline.

        This function can be overridden in a subclass to define custom
        interpolators.

        Parameters
        ----------
        **kwargs : dict, optional
            Additional optional keyword arguments:

            - **degree** : int, tuple, optional
                Degree of the interpolating spline. A tuple can be used
                to provide different degrees for the X- and Y-axes.
                Default value is degree=3.

            - **s** : float, optional
                Non-negative smoothing factor. Default value s=0
                corresponds to interpolation.  See
                :py:class:`~scipy.interpolate.RectBivariateSpline` for
                more details.

        Notes
        -----
        * When subclassing :py:class:`FittableImageModel` for the
          purpose of overriding :py:func:`compute_interpolator`, the
          :py:func:`evaluate` may need to overridden as well depending
          on the behavior of the new interpolator. In addition, for
          improved future compatibility, make sure that the overriding
          method stores keyword arguments ``kwargs`` by calling
          ``_store_interpolator_kwargs`` method.

        * Use caution when modifying interpolator's degree or smoothness
          in a computationally intensive part of the code as it may
          decrease code performance due to the need to recompute
          interpolator.
        """
        from scipy.interpolate import RectBivariateSpline

        if 'degree' in kwargs:
            degree = kwargs['degree']
            if hasattr(degree, '__iter__') and len(degree) == 2:
                degx = int(degree[0])
                degy = int(degree[1])
            else:
                degx = int(degree)
                degy = int(degree)
            if degx < 0 or degy < 0:
                raise ValueError("Interpolator degree must be a non-negative "
                                 "integer")
        else:
            degx = 3
            degy = 3

        if 's' in kwargs:
            smoothness = kwargs['s']
        else:
            smoothness = 0

        # Interpolator must be set to interpolate on the undersampled
        # pixel grid, going from 0 to len(undersampled_grid)
        x = np.arange(self._nx, dtype=float) / self.oversampling[0]
        y = np.arange(self._ny, dtype=float) / self.oversampling[1]
        self.interpolator = RectBivariateSpline(
            x, y, self._data.T, kx=degx, ky=degy, s=smoothness)

        self._store_interpolator_kwargs(**kwargs)

    def evaluate(self, x, y, flux, x_0, y_0):
        """
        Evaluate the model on some input variables and provided model
        parameters.
        """
        xi = np.asarray(x) - x_0 + self._x_origin
        yi = np.asarray(y) - y_0 + self._y_origin

        evaluated_model = flux * self.interpolator.ev(xi, yi)

        if self._fill_value is not None:
            # find indices of pixels that are outside the input pixel
            # grid and set these pixels to the 'fill_value':
            invalid = (((xi < 0) | (xi > (self._nx - 1)
                                    / self.oversampling[0])) |
                       ((yi < 0) | (yi > (self._ny - 1)
                                    / self.oversampling[1])))
            evaluated_model[invalid] = self._fill_value

        return evaluated_model


class GriddedPSFModel(Fittable2DModel):
    """
    A fittable 2D model containing a grid PSF models defined at specific
    locations that are interpolated to evaluate a PSF at an arbitrary
    (x, y) position.

    Parameters
    ----------
    data : `~astropy.nddata.NDData`
        An `~astropy.nddata.NDData` object containing the grid of
        reference PSF arrays.  The data attribute must contain a 3D
        `~numpy.ndarray` containing a stack of the 2D PSFs (the data
        shape should be (N_psf, PSF_ny, PSF_nx)).  The meta
        attribute must be `dict` containing the following:

            * ``'grid_xypos'``:  A list of the (x, y) grid positions of
              each reference PSF.  The order of positions should match
              the first axis of the 3D `~numpy.ndarray` of PSFs.  In
              other words, ``grid_xypos[i]`` should be the (x, y)
              position of the reference PSF defined in ``data[i]``.
            * ``'oversampling'``:  The integer oversampling factor of the
               PSF.

        The meta attribute may contain other properties such as the
        telescope, instrument, detector, and filter of the PSF.
    """

    flux = Parameter(description='Intensity scaling factor for the PSF '
                     'model.', default=1.0)
    x_0 = Parameter(description='x position in the output coordinate grid '
                    'where the model is evaluated.', default=0.0)
    y_0 = Parameter(description='y position in the output coordinate grid '
                    'where the model is evaluated.', default=0.0)

    def __init__(self, data, flux=flux.default, x_0=x_0.default,
                 y_0=y_0.default, fill_value=0.0):

        if not isinstance(data, NDData):
            raise TypeError('data must be an NDData instance.')

        if data.data.ndim != 3:
            raise ValueError('The NDData data attribute must be a 3D numpy '
                             'ndarray')

        if 'grid_xypos' not in data.meta:
            raise ValueError('"grid_xypos" must be in the nddata meta '
                             'dictionary.')
        if len(data.meta['grid_xypos']) != data.data.shape[0]:
            raise ValueError('The length of grid_xypos must match the number '
                             'of input PSFs.')

        if 'oversampling' not in data.meta:
            raise ValueError('"oversampling" must be in the nddata meta '
                             'dictionary.')
        if not np.isscalar(data.meta['oversampling']):
            raise ValueError('oversampling must be a scalar value')

        self.data = np.array(data.data, copy=True, dtype=float)
        self.meta = data.meta
        self.grid_xypos = data.meta['grid_xypos']
        self.oversampling = data.meta['oversampling']

        self._grid_xpos, self._grid_ypos = np.transpose(self.grid_xypos)
        self._xgrid = np.unique(self._grid_xpos)  # also sorts values
        self._ygrid = np.unique(self._grid_ypos)  # also sorts values

        if (len(list(itertools.product(self._xgrid, self._ygrid))) !=
                len(self.grid_xypos)):
            raise ValueError('"grid_xypos" must form a regular grid.')

        self._xgrid_min = self._xgrid[0]
        self._xgrid_max = self._xgrid[-1]
        self._ygrid_min = self._ygrid[0]
        self._ygrid_max = self._ygrid[-1]

        super().__init__(flux, x_0, y_0)

    @staticmethod
    def _find_bounds_1d(data, x):
        """
        Find the index of the lower bound where ``x`` should be inserted
        into ``a`` to maintain order.

        The index of the upper bound is the index of the lower bound
        plus 2.  Both bound indices must be within the array.

        Parameters
        ----------
        data : 1D `~numpy.ndarray`
            The 1D array to search.

        x : float
            The value to insert.

        Returns
        -------
        index : int
            The index of the lower bound.
        """
        idx = np.searchsorted(data, x)
        if idx == 0:
            idx0 = 0
        elif idx == len(data):  # pragma: no cover
            idx0 = idx - 2
        else:
            idx0 = idx - 1

        return idx0

    def _find_bounding_points(self, x, y):
        """
        Find the indices of the grid points that bound the input
        ``(x, y)`` position.

        Parameters
        ----------
        x, y : float
            The ``(x, y)`` position where the PSF is to be evaluated.

        Returns
        -------
        indices : list of int
            A list of indices of the bounding grid points.
        """
        if not np.isscalar(x) or not np.isscalar(y):  # pragma: no cover
            raise TypeError('x and y must be scalars')

        if (x < self._xgrid_min or x > self._xgrid_max or
                y < self._ygrid_min or y > self._ygrid_max):  # pragma: no cover
            raise ValueError('(x, y) position is outside of the region '
                             'defined by grid of PSF positions')

        x0 = self._find_bounds_1d(self._xgrid, x)
        y0 = self._find_bounds_1d(self._ygrid, y)
        points = list(itertools.product(self._xgrid[x0:x0 + 2],
                                        self._ygrid[y0:y0 + 2]))

        indices = []
        for xx, yy in points:
            indices.append(np.argsort(np.hypot(self._grid_xpos - xx,
                                               self._grid_ypos - yy))[0])

        return indices

    @staticmethod
    def _bilinear_interp(xyref, zref, xi, yi):
        """
        Perform bilinear interpolation of four 2D arrays located at
        points on a regular grid.

        Parameters
        ----------
        xyref : list of 4 (x, y) pairs
            A list of 4 ``(x, y)`` pairs that form a rectangle.

        zref : 3D `~numpy.ndarray`
            A 3D `~numpy.ndarray` of shape ``(4, nx, ny)``. The first
            axis corresponds to ``xyref``, i.e., ``refdata[0, :, :]`` is
            the 2D array located at ``xyref[0]``.

        xi, yi : float
            The ``(xi, yi)`` point at which to perform the
            interpolation.  The ``(xi, yi)`` point must lie within the
            rectangle defined by ``xyref``.

        Returns
        -------
        result : 2D `~numpy.ndarray`
            The 2D interpolated array.
        """
        if len(xyref) != 4:
            raise ValueError('xyref must contain only 4 (x, y) pairs')

        if zref.shape[0] != 4:
            raise ValueError('zref must have a length of 4 on the first '
                             'axis.')

        xyref = [tuple(i) for i in xyref]
        idx = sorted(range(len(xyref)), key=xyref.__getitem__)
        xyref = sorted(xyref)  # sort by x, then y
        (x0, y0), (_x0, y1), (x1, _y0), (_x1, _y1) = xyref

        if x0 != _x0 or x1 != _x1 or y0 != _y0 or y1 != _y1:
            raise ValueError('The refxy points do not form a rectangle.')

        if not np.isscalar(xi):
            xi = xi[0]
        if not np.isscalar(yi):
            yi = yi[0]

        if not x0 <= xi <= x1 or not y0 <= yi <= y1:
            raise ValueError('The (x, y) input is not within the rectangle '
                             'defined by xyref.')

        data = np.asarray(zref)[idx]
        weights = np.array([(x1 - xi) * (y1 - yi), (x1 - xi) * (yi - y0),
                            (xi - x0) * (y1 - yi), (xi - x0) * (yi - y0)])
        norm = (x1 - x0) * (y1 - y0)

        return np.sum(data * weights[:, None, None], axis=0) / norm

    def _compute_local_model(self, x_0, y_0):
        """
        Return `FittableImageModel` for interpolated PSF at some (x_0, y_0).
        """
        # NOTE: this is needed because the PSF photometry routines input
        # length-1 values instead of scalars.  TODO: fix the photometry
        # routines.
        if not np.isscalar(x_0):
            x_0 = x_0[0]
        if not np.isscalar(y_0):
            y_0 = y_0[0]

        if (x_0 < self._xgrid_min or x_0 > self._xgrid_max or
                y_0 < self._ygrid_min or y_0 > self._ygrid_max):

            # position is outside of the grid, so simply use the
            # closest reference PSF
            self._ref_indices = np.argsort(np.hypot(self._grid_xpos - x_0,
                                                    self._grid_ypos - y_0))[0]
            self._psf_interp = self.data[self._ref_indices, :, :]
        else:
            # find the four bounding reference PSFs and interpolate
            self._ref_indices = self._find_bounding_points(x_0, y_0)
            xyref = np.array(self.grid_xypos)[self._ref_indices]
            psfs = self.data[self._ref_indices, :, :]

            self._psf_interp = self._bilinear_interp(xyref, psfs, x_0, y_0)

        # Construct the model using the interpolated supersampled data
        psfmodel = FittableImageModel(self._psf_interp,
                                      oversampling=self.oversampling)
        return psfmodel

    def evaluate(self, x, y, flux, x_0, y_0):
        """
        Evaluate the `GriddedPSFModel` for the input parameters.
        """
        # Get the local PSF at the (x_0,y_0)
        psfmodel = self._compute_local_model(x_0, y_0)

        # now evaluate the PSF at the (x_0, y_0) subpixel position on
        # the input (x, y) values
        return psfmodel.evaluate(x, y, flux, x_0, y_0)


class IntegratedGaussianPRF(Fittable2DModel):
    r"""
    Circular Gaussian model integrated over pixels.

    Because it is integrated, this model is considered a PRF, *not* a
    PSF (see :ref:`psf-terminology` for more about the terminology used
    here.)

    This model is a Gaussian *integrated* over an area of
    ``1`` (in units of the model input coordinates, e.g., 1
    pixel). This is in contrast to the apparently similar
    `astropy.modeling.functional_models.Gaussian2D`, which is the value
    of a 2D Gaussian *at* the input coordinates, with no integration.
    So this model is equivalent to assuming the PSF is Gaussian at a
    *sub-pixel* level.

    Parameters
    ----------
    sigma : float
        Width of the Gaussian PSF.
    flux : float, optional
        Total integrated flux over the entire PSF
    x_0 : float, optional
        Position of the peak in x direction.
    y_0 : float, optional
        Position of the peak in y direction.

    Notes
    -----
    This model is evaluated according to the following formula:

        .. math::

            f(x, y) =
                \frac{F}{4}
                \left[
                {\rm erf} \left(\frac{x - x_0 + 0.5}
                {\sqrt{2} \sigma} \right) -
                {\rm erf} \left(\frac{x - x_0 - 0.5}
                {\sqrt{2} \sigma} \right)
                \right]
                \left[
                {\rm erf} \left(\frac{y - y_0 + 0.5}
                {\sqrt{2} \sigma} \right) -
                {\rm erf} \left(\frac{y - y_0 - 0.5}
                {\sqrt{2} \sigma} \right)
                \right]

    where ``erf`` denotes the error function and ``F`` the total
    integrated flux.
    """

    flux = Parameter(default=1)
    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)
    sigma = Parameter(default=1, fixed=True)

    _erf = None

    @property
    def bounding_box(self):
        halfwidth = 4 * self.sigma
        return ((int(self.y_0 - halfwidth), int(self.y_0 + halfwidth)),
                (int(self.x_0 - halfwidth), int(self.x_0 + halfwidth)))

    def __init__(self, sigma=sigma.default,
                 x_0=x_0.default, y_0=y_0.default, flux=flux.default,
                 **kwargs):
        if self._erf is None:
            from scipy.special import erf
            self.__class__._erf = erf

        super().__init__(n_models=1, sigma=sigma, x_0=x_0, y_0=y_0, flux=flux,
                         **kwargs)

    def evaluate(self, x, y, flux, x_0, y_0, sigma):
        """Model function Gaussian PSF model."""
        return (flux / 4 *
                ((self._erf((x - x_0 + 0.5) / (np.sqrt(2) * sigma)) -
                  self._erf((x - x_0 - 0.5) / (np.sqrt(2) * sigma))) *
                 (self._erf((y - y_0 + 0.5) / (np.sqrt(2) * sigma)) -
                  self._erf((y - y_0 - 0.5) / (np.sqrt(2) * sigma)))))


class PRFAdapter(Fittable2DModel):
    """
    A model that adapts a supplied PSF model to act as a PRF.

    It integrates the PSF model over pixel "boxes".  A critical built-in
    assumption is that the PSF model scale and location parameters are
    in *pixel* units.

    Parameters
    ----------
    psfmodel : a 2D model
        The model to assume as representative of the PSF
    renormalize_psf : bool
        If True, the model will be integrated from -inf to inf and
        re-scaled so that the total integrates to 1.  Note that this
        renormalization only occurs *once*, so if the total flux of
        ``psfmodel`` depends on position, this will *not* be correct.
    xname : str or None
        The name of the ``psfmodel`` parameter that corresponds to the
        x-axis center of the PSF.  If None, the model will be assumed to
        be centered at x=0.
    yname : str or None
        The name of the ``psfmodel`` parameter that corresponds to the
        y-axis center of the PSF.  If None, the model will be assumed to
        be centered at y=0.
    fluxname : str or None
        The name of the ``psfmodel`` parameter that corresponds to the
        total flux of the star.  If None, a scaling factor will be
        applied by the ``PRFAdapter`` instead of modifying the
        ``psfmodel``.

    Notes
    -----
    This current implementation of this class (using numerical
    integration for each pixel) is extremely slow, and only suited for
    experimentation over relatively few small regions.
    """

    flux = Parameter(default=1)
    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)

    def __init__(self, psfmodel, renormalize_psf=True, flux=flux.default,
                 x_0=x_0.default, y_0=y_0.default, xname=None, yname=None,
                 fluxname=None, **kwargs):

        self.psfmodel = psfmodel.copy()

        if renormalize_psf:
            from scipy.integrate import dblquad
            self._psf_scale_factor = 1. / dblquad(self.psfmodel,
                                                  -np.inf, np.inf,
                                                  lambda x: -np.inf,
                                                  lambda x: np.inf)[0]
        else:
            self._psf_scale_factor = 1

        self.xname = xname
        self.yname = yname
        self.fluxname = fluxname

        # these can be used to adjust the integration behavior. Might be
        # used in the future to expose how the integration happens
        self._dblquadkwargs = {}

        super().__init__(n_models=1, x_0=x_0, y_0=y_0, flux=flux, **kwargs)

    def evaluate(self, x, y, flux, x_0, y_0):
        """The evaluation function for PRFAdapter."""
        if self.xname is None:
            dx = x - x_0
        else:
            dx = x
            setattr(self.psfmodel, self.xname, x_0)

        if self.xname is None:
            dy = y - y_0
        else:
            dy = y
            setattr(self.psfmodel, self.yname, y_0)

        if self.fluxname is None:
            return (flux * self._psf_scale_factor *
                    self._integrated_psfmodel(dx, dy))
        else:
            setattr(self.psfmodel, self.yname, flux * self._psf_scale_factor)
            return self._integrated_psfmodel(dx, dy)

    def _integrated_psfmodel(self, dx, dy):
        from scipy.integrate import dblquad

        # infer type/shape from the PSF model.  Seems wasteful, but the
        # integration step is a *lot* more expensive so its just peanuts
        out = np.empty_like(self.psfmodel(dx, dy))
        outravel = out.ravel()
        for i, (xi, yi) in enumerate(zip(dx.ravel(), dy.ravel())):
            outravel[i] = dblquad(self.psfmodel,
                                  xi-0.5, xi+0.5,
                                  lambda x: yi-0.5, lambda x: yi+0.5,
                                  **self._dblquadkwargs)[0]
        return out



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the sandbox module.
"""

from astropy.convolution.utils import discretize_model
from astropy.modeling.models import Gaussian2D
from astropy.table import Table
import numpy as np
from numpy.testing import assert_allclose

from ..sandbox import DiscretePRF


PSF_SIZE = 11
GAUSSIAN_WIDTH = 1.
IMAGE_SIZE = 101

# Position and FLUXES of test sources
INTAB = Table([[50., 23, 12, 86], [50., 83, 80, 84],
               [np.pi * 10, 3.654, 20., 80 / np.sqrt(3)]],
              names=['x_0', 'y_0', 'flux_0'])

# Create test psf
psf_model = Gaussian2D(1. / (2 * np.pi * GAUSSIAN_WIDTH ** 2), PSF_SIZE // 2,
                       PSF_SIZE // 2, GAUSSIAN_WIDTH, GAUSSIAN_WIDTH)
test_psf = discretize_model(psf_model, (0, PSF_SIZE), (0, PSF_SIZE),
                            mode='oversample')

# Set up grid for test image
image = np.zeros((IMAGE_SIZE, IMAGE_SIZE))

# Add sources to test image
for x, y, flux in INTAB:
    model = Gaussian2D(flux / (2 * np.pi * GAUSSIAN_WIDTH ** 2),
                       x, y, GAUSSIAN_WIDTH, GAUSSIAN_WIDTH)
    image += discretize_model(model, (0, IMAGE_SIZE), (0, IMAGE_SIZE),
                              mode='oversample')

# Some tests require an image with wider sources.
WIDE_GAUSSIAN_WIDTH = 3.
WIDE_INTAB = Table([[50, 23.2], [50.5, 1], [10, 20]],
                   names=['x_0', 'y_0', 'flux_0'])
wide_image = np.zeros((IMAGE_SIZE, IMAGE_SIZE))

# Add sources to test image
for x, y, flux in WIDE_INTAB:
    model = Gaussian2D(flux / (2 * np.pi * WIDE_GAUSSIAN_WIDTH ** 2),
                       x, y, WIDE_GAUSSIAN_WIDTH, WIDE_GAUSSIAN_WIDTH)
    wide_image += discretize_model(model, (0, IMAGE_SIZE), (0, IMAGE_SIZE),
                                   mode='oversample')


def test_create_prf_mean():
    """
    Check if create_prf works correctly on simulated data.
    Position input format: list
    """

    prf = DiscretePRF.create_from_image(image,
                                        list(INTAB['x_0', 'y_0'].as_array()),
                                        PSF_SIZE, subsampling=1, mode='mean')
    assert_allclose(prf._prf_array[0, 0], test_psf, atol=1E-8)


def test_create_prf_median():
    """
    Check if create_prf works correctly on simulated data.
    Position input format: astropy.table.Table
    """

    prf = DiscretePRF.create_from_image(image, np.array(INTAB['x_0', 'y_0']),
                                        PSF_SIZE, subsampling=1,
                                        mode='median')
    assert_allclose(prf._prf_array[0, 0], test_psf, atol=1E-8)


def test_create_prf_nan():
    """
    Check if create_prf deals correctly with nan values.
    """

    image_nan = image.copy()
    image_nan[52, 52] = np.nan
    image_nan[52, 48] = np.nan
    prf = DiscretePRF.create_from_image(image, np.array(INTAB['x_0', 'y_0']),
                                        PSF_SIZE, subsampling=1, fix_nan=True)
    assert not np.isnan(prf._prf_array[0, 0]).any()


def test_create_prf_flux():
    """
    Check if create_prf works correctly when FLUXES are specified.
    """

    prf = DiscretePRF.create_from_image(image, np.array(INTAB['x_0', 'y_0']),
                                        PSF_SIZE, subsampling=1,
                                        mode='median', fluxes=INTAB['flux_0'])
    assert_allclose(prf._prf_array[0, 0].sum(), 1)
    assert_allclose(prf._prf_array[0, 0], test_psf, atol=1E-8)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage provides general-purpose utility functions.
"""

from .colormaps import *  # noqa
from .errors import *  # noqa
from .exceptions import *  # noqa
from .interpolation import *  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides functions for measuring non-parametric morphologies
of sources.
"""

import numpy as np

__all__ = ['gini']


def gini(data):
    r"""
    Calculate the `Gini coefficient
    <https://en.wikipedia.org/wiki/Gini_coefficient>`_ of a 2D array.

    The Gini coefficient is calculated using the prescription from `Lotz
    et al. 2004
    <https://ui.adsabs.harvard.edu/abs/2004AJ....128..163L/abstract>`_
    as:

    .. math::
        G = \frac{1}{\left | \bar{x} \right | n (n - 1)}
        \sum^{n}_{i} (2i - n - 1) \left | x_i \right |

    where :math:`\bar{x}` is the mean over all pixel values
    :math:`x_i`.

    The Gini coefficient is a way of measuring the inequality in a given
    set of values. In the context of galaxy morphology, it measures how
    the light of a galaxy image is distributed among its pixels.  A Gini
    coefficient value of 0 corresponds to a galaxy image with the light
    evenly distributed over all pixels while a Gini coefficient value of
    1 represents a galaxy image with all its light concentrated in just
    one pixel.

    Usually Gini's measurement needs some sort of preprocessing for
    defining the galaxy region in the image based on the quality of the
    input data. As there is not a general standard for doing this, this
    is left for the user.

    Parameters
    ----------
    data : array-like
        The 2D data array or object that can be converted to an array.

    Returns
    -------
    gini : `float`
        The Gini coefficient of the input 2D array.
    """
    flattened = np.sort(np.ravel(data))
    npix = np.size(flattened)
    normalization = np.abs(np.mean(flattened)) * npix * (npix - 1)
    kernel = (2. * np.arange(1, npix + 1) - npix - 1) * np.abs(flattened)

    return np.sum(kernel) / normalization



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the errors module.
"""

import astropy.units as u
import numpy as np
from numpy.testing import assert_allclose
import pytest

from ..errors import calc_total_error


SHAPE = (5, 5)
DATAVAL = 2.
DATA = np.ones(SHAPE) * DATAVAL
BKG_ERROR = np.ones(SHAPE)
EFFGAIN = np.ones(SHAPE) * DATAVAL
BACKGROUND = np.ones(SHAPE)
WRONG_SHAPE = np.ones((2, 2))


def test_error_shape():
    with pytest.raises(ValueError):
        calc_total_error(DATA, WRONG_SHAPE, EFFGAIN)


def test_gain_shape():
    with pytest.raises(ValueError):
        calc_total_error(DATA, BKG_ERROR, WRONG_SHAPE)


@pytest.mark.parametrize('effective_gain', (-1, -100))
def test_gain_negative(effective_gain):
    with pytest.raises(ValueError):
        calc_total_error(DATA, BKG_ERROR, effective_gain)


def test_gain_scalar():
    error_tot = calc_total_error(DATA, BKG_ERROR, 2.)
    assert_allclose(error_tot, np.sqrt(2.) * BKG_ERROR)


def test_gain_array():
    error_tot = calc_total_error(DATA, BKG_ERROR, EFFGAIN)
    assert_allclose(error_tot, np.sqrt(2.) * BKG_ERROR)


def test_gain_zero():
    error_tot = calc_total_error(DATA, BKG_ERROR, 0.)
    assert_allclose(error_tot, BKG_ERROR)

    effgain = np.copy(EFFGAIN)
    effgain[0, 0] = 0
    effgain[1, 1] = 0
    mask = (effgain == 0)
    error_tot = calc_total_error(DATA, BKG_ERROR, effgain)
    assert_allclose(error_tot[mask], BKG_ERROR[mask])
    assert_allclose(error_tot[~mask], np.sqrt(2))


def test_units():
    units = u.electron / u.s
    error_tot1 = calc_total_error(DATA * units, BKG_ERROR * units,
                                  EFFGAIN * u.s)
    assert error_tot1.unit == units
    error_tot2 = calc_total_error(DATA, BKG_ERROR, EFFGAIN)
    assert_allclose(error_tot1.value, error_tot2)


def test_error_units():
    units = u.electron / u.s
    with pytest.raises(ValueError):
        calc_total_error(DATA * units, BKG_ERROR * u.electron,
                         EFFGAIN * u.s)


def test_effgain_units():
    units = u.electron / u.s
    with pytest.raises(u.UnitsError):
        calc_total_error(DATA * units, BKG_ERROR * units, EFFGAIN * u.km)


def test_missing_bkgerror_units():
    units = u.electron / u.s
    with pytest.raises(ValueError):
        calc_total_error(DATA * units, BKG_ERROR, EFFGAIN * u.s)


def test_missing_effgain_units():
    units = u.electron / u.s
    with pytest.raises(ValueError):
        calc_total_error(DATA * units, BKG_ERROR * units,
                         EFFGAIN)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the properties module.
"""

import itertools

import astropy.units as u
import astropy.wcs as WCS
from astropy.modeling import models
from astropy.table import QTable
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.exceptions import AstropyDeprecationWarning
from astropy.utils.misc import isiterable
from numpy.testing import assert_allclose
import numpy as np
import pytest

from ..core import SegmentationImage
from ..detect import detect_sources
from ..properties import (LegacySourceCatalog, SourceProperties,
                          source_properties)
from ...datasets import make_gwcs, make_wcs
from ...utils._optional_deps import HAS_GWCS, HAS_SCIPY  # noqa


XCEN = 51.
YCEN = 52.7
MAJOR_SIG = 8.
MINOR_SIG = 3.
THETA = np.pi / 6.
G1 = models.Gaussian2D(111., XCEN, YCEN, MAJOR_SIG, MINOR_SIG, theta=THETA)
G2 = models.Gaussian2D(50, 20, 80, 5.1, 4.5)
G3 = models.Gaussian2D(70, 75, 18, 9.2, 4.5)
Y, X = np.mgrid[0:100, 0:100]
IMAGE = G1(X, Y) + G2(X, Y) + G3(X, Y)
THRESHOLD = 0.1
ERR_VALS = [0., 2.5]
BACKGRD_VALS = [None, 0., 1., 3.5]
FWHM2SIGMA = 1.0 / (2.0 * np.sqrt(2.0 * np.log(2.0)))


@pytest.mark.skipif('not HAS_SCIPY')
class TestSourceProperties:
    def setup_class(self):
        self.segm = detect_sources(IMAGE, THRESHOLD, npixels=5)

    def test_invalid_shapes(self):
        with pytest.warns(AstropyDeprecationWarning):
            wrong_shape = np.ones((3, 3))
            with pytest.raises(ValueError):
                SourceProperties(IMAGE, np.eye(3, dtype=int), label=1)

            with pytest.raises(ValueError):
                SourceProperties(IMAGE, self.segm, label=1,
                                 filtered_data=wrong_shape)

            with pytest.raises(ValueError):
                SourceProperties(IMAGE, self.segm, label=1, error=wrong_shape)

            with pytest.raises(ValueError):
                SourceProperties(IMAGE, self.segm, label=1,
                                 background=wrong_shape)

    def test_invalid_units(self):
        with pytest.warns(AstropyDeprecationWarning):
            unit = u.uJy
            wrong_unit = u.km

            with pytest.raises(ValueError):
                SourceProperties(IMAGE*unit, self.segm, label=1,
                                 filtered_data=IMAGE*wrong_unit)

            with pytest.raises(ValueError):
                SourceProperties(IMAGE*unit, self.segm, label=1,
                                 error=IMAGE*wrong_unit)

            with pytest.raises(ValueError):
                SourceProperties(IMAGE*unit, self.segm, label=1,
                                 background=IMAGE*wrong_unit)

            # all array inputs must have the same unit
            with pytest.raises(ValueError):
                SourceProperties(IMAGE*unit, self.segm, label=1,
                                 filtered_data=IMAGE)

    @pytest.mark.parametrize('label', (0, -1))
    def test_label_invalid(self, label):
        with pytest.warns(AstropyDeprecationWarning):
            with pytest.raises(ValueError):
                SourceProperties(IMAGE, self.segm, label=label)

    @pytest.mark.parametrize('label', (0, -1))
    def test_label_missing(self, label):
        with pytest.warns(AstropyDeprecationWarning):
            segm = self.segm.copy()
            segm.remove_label(2)
            with pytest.raises(ValueError):
                SourceProperties(IMAGE, segm, label=2)
                SourceProperties(IMAGE, segm, label=label)

    def test_wcs(self):
        with pytest.warns(AstropyDeprecationWarning):
            mywcs = make_wcs(IMAGE.shape)
            props = SourceProperties(IMAGE, self.segm, wcs=mywcs, label=1)
            assert props.sky_centroid is not None
            assert props.sky_centroid_icrs is not None
            assert props.sky_bbox_ll is not None
            assert props.sky_bbox_ul is not None
            assert props.sky_bbox_lr is not None
            assert props.sky_bbox_ur is not None

            tbl = props.to_table()
            assert len(tbl) == 1

    @pytest.mark.skipif('not HAS_GWCS')
    def test_gwcs(self):
        with pytest.warns(AstropyDeprecationWarning):
            mywcs = make_gwcs(IMAGE.shape)
            props = SourceProperties(IMAGE, self.segm, wcs=mywcs, label=1)
            assert props.sky_centroid is not None
            assert props.sky_centroid_icrs is not None
            assert props.sky_bbox_ll is not None
            assert props.sky_bbox_ul is not None
            assert props.sky_bbox_lr is not None
            assert props.sky_bbox_ur is not None

            tbl = props.to_table()
            assert len(tbl) == 1

    def test_nowcs(self):
        with pytest.warns(AstropyDeprecationWarning):
            props = SourceProperties(IMAGE, self.segm, wcs=None, label=1)
            assert props.sky_centroid_icrs is None
            assert props.sky_bbox_ll is None
            assert props.sky_bbox_ul is None
            assert props.sky_bbox_lr is None
            assert props.sky_bbox_ur is None

    def test_to_table(self):
        with pytest.warns(AstropyDeprecationWarning):
            props = SourceProperties(IMAGE, self.segm, label=2)
            t1 = props.to_table()
            assert isinstance(t1, QTable)
            assert len(t1) == 1
            assert_quantity_allclose(t1['area'], 1058 * u.pix**2)

    def test_masks(self):
        """
        Test masks, including automatic masking of all non-finite (e.g.,
        NaN, inf) values in the data array.
        """
        with pytest.warns(AstropyDeprecationWarning):
            error = np.ones(IMAGE.shape) * 5.1
            error[41, 35] = np.nan
            error[42, 36] = np.inf
            background = np.ones(IMAGE.shape) * 1.2
            background[62, 55] = np.nan
            background[63, 56] = np.inf
            mask = np.zeros(IMAGE.shape).astype(bool)
            mask[45:55, :] = True
            data = np.copy(IMAGE)
            data[40, 40:45] = np.nan
            data[60, 60:65] = np.inf
            data[65, 65:70] = -np.inf

            props = SourceProperties(data, self.segm, label=2, error=error,
                                     background=background, mask=mask)

            # ensure mask is identical for data, error, and background
            assert props.data_cutout_ma.compressed().size == 677
            assert (props.data_cutout_ma.compressed().size ==
                    props.filtered_data_cutout_ma.compressed().size)
            assert (props.data_cutout_ma.compressed().size ==
                    props.error_cutout_ma.compressed().size)
            assert (props.data_cutout_ma.compressed().size ==
                    props.background_cutout_ma.compressed().size)
            assert (len(props._filtered_data_values) ==
                    props.filtered_data_cutout_ma.compressed().size)

            # test for non-finite values in error and/or background outside
            # of the data mask
            tbl = props.to_table()
            assert np.isnan(tbl['source_sum_err'])
            assert np.isnan(tbl['background_sum'])
            assert np.isnan(tbl['background_mean'])

            # test that the masks are independent objects
            assert (np.count_nonzero(props._segment_mask) !=
                    np.count_nonzero(props._total_mask))
            assert (np.count_nonzero(props._data_mask) !=
                    np.count_nonzero(props._total_mask))

            assert_allclose(props._data_zeroed.sum(), props.source_sum)

            assert_allclose(props.data_cutout,
                            props.make_cutout(props._data, masked_array=False))
            assert_allclose(props.data_cutout_ma,
                            props.make_cutout(props._data, masked_array=True))
            assert_allclose(props.error_cutout_ma,
                            props.make_cutout(props._error, masked_array=True))
            assert_allclose(props.background_cutout_ma,
                            props.make_cutout(props._background,
                                              masked_array=True))

    def test_completely_masked(self):
        """Test case where a source is completely masked."""
        with pytest.warns(AstropyDeprecationWarning):
            error = np.ones(IMAGE.shape) * 5.1
            background = np.ones(IMAGE.shape) * 1.2
            mask = np.ones(IMAGE.shape).astype(bool)
            obj = source_properties(IMAGE, self.segm, error=error,
                                    background=background, mask=mask)[0]
            assert np.isnan(obj.xcentroid.value)
            assert np.isnan(obj.ycentroid.value)
            assert np.isnan(obj.source_sum)
            assert np.isnan(obj.source_sum_err)
            assert np.isnan(obj.background_sum)
            assert np.isnan(obj.background_mean)
            assert np.isnan(obj.background_at_centroid)
            assert np.isnan(obj.area)
            assert np.isnan(obj.perimeter)
            assert np.isnan(obj.min_value)
            assert np.isnan(obj.max_value)
            assert np.isnan(obj.minval_xpos.value)
            assert np.isnan(obj.minval_ypos.value)
            assert np.isnan(obj.maxval_xpos.value)
            assert np.isnan(obj.maxval_ypos.value)
            assert np.all(np.isnan(obj.minval_cutout_pos.value))
            assert np.all(np.isnan(obj.maxval_cutout_pos.value))
            assert np.isnan(obj.gini)

    def test_repr_str(self):
        with pytest.warns(AstropyDeprecationWarning):
            props = SourceProperties(IMAGE, self.segm, label=1)
            assert repr(props) == str(props)

            attrs = ['label', 'centroid', 'sky_centroid']
            for attr in attrs:
                assert f'{attr}:' in repr(props)


@pytest.mark.skipif('not HAS_SCIPY')
class TestSourcePropertiesFunctionInputs:
    def setup_class(self):
        self.segm = detect_sources(IMAGE, THRESHOLD, npixels=5)

    def test_segment_shape(self):
        with pytest.warns(AstropyDeprecationWarning):
            wrong_shape = np.eye(3, dtype=int)
            with pytest.raises(ValueError):
                source_properties(IMAGE, wrong_shape)

    def test_error_shape(self):
        with pytest.warns(AstropyDeprecationWarning):
            wrong_shape = np.ones((2, 2))
            with pytest.raises(ValueError):
                source_properties(IMAGE, self.segm, error=wrong_shape)

    def test_background_shape(self):
        with pytest.warns(AstropyDeprecationWarning):
            wrong_shape = np.ones((2, 2))
            with pytest.raises(ValueError):
                source_properties(IMAGE, self.segm, background=wrong_shape)

    def test_mask_shape(self):
        with pytest.warns(AstropyDeprecationWarning):
            wrong_shape = np.zeros((2, 2)).astype(bool)
            with pytest.raises(ValueError):
                source_properties(IMAGE, self.segm, mask=wrong_shape)

    def test_labels(self):
        with pytest.warns(AstropyDeprecationWarning):
            props = source_properties(IMAGE, self.segm, labels=1)
            assert props[0].id == 1

    def test_nosources(self):
        with pytest.warns(AstropyDeprecationWarning):
            with pytest.raises(ValueError):
                source_properties(IMAGE, self.segm, labels=-1)


@pytest.mark.skipif('not HAS_SCIPY')
class TestSourcePropertiesFunction:
    def setup_class(self):
        self.segm = detect_sources(IMAGE, THRESHOLD, npixels=5)

    def test_properties(self):
        with pytest.warns(AstropyDeprecationWarning):
            obj = source_properties(IMAGE, self.segm)[1]
            assert obj.id == 2
            assert_quantity_allclose(obj.xcentroid, XCEN*u.pix,
                                     rtol=1.e-2)
            assert_quantity_allclose(obj.ycentroid, YCEN*u.pix,
                                     rtol=1.e-2)
            assert_allclose(obj.source_sum, 16723.388)
            assert_quantity_allclose(obj.semimajor_axis_sigma,
                                     MAJOR_SIG*u.pix, rtol=1.e-2)
            assert_quantity_allclose(obj.semiminor_axis_sigma,
                                     MINOR_SIG*u.pix, rtol=1.e-2)
            assert_quantity_allclose(obj.orientation, THETA*u.rad,
                                     rtol=1.e-3)
            assert obj.bbox_xmin.value == 25
            assert obj.bbox_xmax.value == 77
            assert obj.bbox_ymin.value == 35
            assert obj.bbox_ymax.value == 70
            assert_quantity_allclose(obj.area, 1058.0*u.pix**2)
            assert_allclose(len(obj.indices), 2)
            assert_allclose(len(obj.indices[0]), obj.area.value)

            properties = ['background_at_centroid', 'background_mean',
                          'eccentricity', 'ellipticity', 'elongation',
                          'equivalent_radius', 'max_value', 'maxval_xpos',
                          'maxval_ypos', 'min_value', 'minval_xpos',
                          'minval_ypos', 'perimeter', 'cxx', 'cxy', 'cyy',
                          'covar_sigx2', 'covar_sigxy', 'covar_sigy2',
                          'bbox_xmin', 'bbox_xmax', 'bbox_ymin', 'bbox_ymax']
            for propname in properties:
                assert not isiterable(getattr(obj, propname))

            properties = ['centroid', 'covariance_eigvals', 'cutout_centroid',
                          'maxval_cutout_pos', 'minval_cutout_pos']
            shapes = [getattr(obj, p).shape for p in properties]
            for shape in shapes:
                assert shape == (2,)

            properties = ['covariance', 'inertia_tensor']
            shapes = [getattr(obj, p).shape for p in properties]
            for shape in shapes:
                assert shape == (2, 2)

            properties = ['moments', 'moments_central']
            shapes = [getattr(obj, p).shape for p in properties]
            for shape in shapes:
                assert shape == (4, 4)

            assert obj.kron_radius.value < 1.3
            assert obj.kron_flux < 16700.
            assert obj.kron_fluxerr is None

    def test_properties_background_notnone(self):
        with pytest.warns(AstropyDeprecationWarning):
            value = 1.
            props = source_properties(IMAGE, self.segm, background=value)
            assert props[0].background_mean == value
            assert_allclose(props[0].background_at_centroid, value)

    def test_properties_background_units(self):
        with pytest.warns(AstropyDeprecationWarning):
            unit = u.uJy
            value = 1. * unit
            props = source_properties(IMAGE * unit, self.segm,
                                      background=value)
            assert props[0].background_mean == value
            assert_allclose(props[0].background_at_centroid, value)

    def test_properties_error_background_none(self):
        with pytest.warns(AstropyDeprecationWarning):
            props = source_properties(IMAGE, self.segm)
            assert props[0].background_cutout_ma is None
            assert props[0].error_cutout_ma is None

    def test_cutout_shapes(self):
        with pytest.warns(AstropyDeprecationWarning):
            error = np.ones(IMAGE.shape, dtype=float)
            props = source_properties(IMAGE, self.segm, error=error,
                                      background=1.)
            bbox = props[0].bbox
            properties = ['background_cutout_ma', 'data_cutout',
                          'data_cutout_ma', 'error_cutout_ma']
            shapes = [getattr(props[0], p).shape for p in properties]
            for shape in shapes:
                assert shape == bbox.shape

    def test_make_cutout(self):
        with pytest.warns(AstropyDeprecationWarning):
            props = source_properties(IMAGE, self.segm)
            data = np.ones((2, 2))
            with pytest.raises(ValueError):
                props[0].make_cutout(data)

    @pytest.mark.parametrize(('error_value', 'background'),
                             list(itertools.product(ERR_VALS, BACKGRD_VALS)))
    def test_segmentation_inputs(self, error_value, background):
        with pytest.warns(AstropyDeprecationWarning):
            error = np.ones(IMAGE.shape) * error_value
            props = source_properties(IMAGE, self.segm, error=error,
                                      background=background)
            obj = props[1]
            assert_quantity_allclose(obj.xcentroid, XCEN*u.pix, rtol=1.e-2)
            assert_quantity_allclose(obj.ycentroid, YCEN*u.pix, rtol=1.e-2)
            assert_quantity_allclose(obj.semimajor_axis_sigma,
                                     MAJOR_SIG*u.pix, rtol=1.e-2)
            assert_quantity_allclose(obj.semiminor_axis_sigma,
                                     MINOR_SIG*u.pix, rtol=1.e-2)
            assert_quantity_allclose(obj.orientation, THETA*u.rad,
                                     rtol=1.e-3)
            assert obj.bbox_xmin.value == 25
            assert obj.bbox_xmax.value == 77
            assert obj.bbox_ymin.value == 35
            assert obj.bbox_ymax.value == 70

            area = obj.area.value
            assert_allclose(area, 1058.0)

            if background is not None:
                assert_allclose(obj.background_sum, area * background)
            assert_allclose(obj.source_sum, 16723.388)

            true_error = np.sqrt(obj.area.value) * error_value
            assert_allclose(obj.source_sum_err, true_error)

    def test_data_allzero(self):
        with pytest.warns(AstropyDeprecationWarning):
            props = source_properties(IMAGE*0., self.segm)
            proplist = ['xcentroid', 'ycentroid', 'semimajor_axis_sigma',
                        'semiminor_axis_sigma', 'eccentricity', 'orientation',
                        'ellipticity', 'elongation', 'cxx', 'cxy', 'cyy']
            for prop in proplist:
                assert np.isnan(getattr(props[0], prop))

    def test_mask(self):
        with pytest.warns(AstropyDeprecationWarning):
            data = np.zeros((3, 3))
            data[0, 1] = 1.
            data[1, 1] = 1.
            mask = np.zeros(data.shape, dtype=bool)
            mask[0, 1] = True
            segm = data.astype(int)
            props = source_properties(data, segm, mask=mask)
            assert_allclose(props[0].xcentroid.value, 1)
            assert_allclose(props[0].ycentroid.value, 1)
            assert_allclose(props[0].source_sum, 1)
            assert_allclose(props[0].area.value, 1)

    def test_mask_nomask(self):
        with pytest.warns(AstropyDeprecationWarning):
            props = source_properties(IMAGE, self.segm, mask=np.ma.nomask)
            mask = np.zeros(IMAGE.shape).astype(bool)
            props2 = source_properties(IMAGE, self.segm, mask=mask)
            assert_allclose(props.xcentroid.value, props2.xcentroid.value)
            assert_allclose(props.ycentroid.value, props2.ycentroid.value)
            assert_quantity_allclose(props.source_sum, props2.source_sum)

    def test_single_pixel_segment(self):
        with pytest.warns(AstropyDeprecationWarning):
            segm = np.zeros(self.segm.shape, dtype=int)
            segm[50, 50] = 1
            props = source_properties(IMAGE, segm)
            assert props[0].eccentricity == 0

    def test_filtering(self):
        with pytest.warns(AstropyDeprecationWarning):
            from astropy.convolution import Gaussian2DKernel
            filter_kernel = Gaussian2DKernel(2.*FWHM2SIGMA, x_size=3, y_size=3)
            error = np.sqrt(IMAGE)
            props1 = source_properties(IMAGE, self.segm, error=error)
            props2 = source_properties(IMAGE, self.segm, error=error,
                                       filter_kernel=filter_kernel.array)
            p1, p2 = props1[0], props2[0]
            keys = ['source_sum', 'source_sum_err']
            for key in keys:
                assert getattr(p1, key) == getattr(p2, key)
            keys = ['semimajor_axis_sigma', 'semiminor_axis_sigma']
            for key in keys:
                assert getattr(p1, key) != getattr(p2, key)

    def test_filtering_kernel(self):
        with pytest.warns(AstropyDeprecationWarning):
            data = np.zeros((3, 3))
            data[1, 1] = 1.
            from astropy.convolution import Gaussian2DKernel
            filter_kernel = Gaussian2DKernel(2.*FWHM2SIGMA, x_size=3, y_size=3)
            error = np.sqrt(IMAGE)
            props1 = source_properties(IMAGE, self.segm, error=error)
            props2 = source_properties(IMAGE, self.segm, error=error,
                                       filter_kernel=filter_kernel)
            p1, p2 = props1[0], props2[0]
            keys = ['source_sum', 'source_sum_err']
            for key in keys:
                assert getattr(p1, key) == getattr(p2, key)
            keys = ['semimajor_axis_sigma', 'semiminor_axis_sigma']
            for key in keys:
                assert getattr(p1, key) != getattr(p2, key)

    def test_data_nan(self):
        """Test case when data contains NaNs within a segment."""
        with pytest.warns(AstropyDeprecationWarning):
            data = np.ones((20, 20))
            data[2, 2] = np.nan
            segm = np.zeros((20, 20)).astype(int)
            segm[1:5, 1:5] = 1
            segm[7:15, 7:15] = 2
            segm = SegmentationImage(segm)
            props = source_properties(data, segm)
            assert_quantity_allclose(props.minval_xpos, [1, 7]*u.pix)

    def test_nonconsecutive_labels(self):
        with pytest.warns(AstropyDeprecationWarning):
            segm = self.segm.copy()
            segm.reassign_label(1, 1000)
            cat = source_properties(IMAGE, segm)
            assert len(cat) == 3
            assert cat.id == [2, 3, 1000]
            assert cat[0].id == 2
            assert cat[1].id == 3
            assert cat[segm.get_index(label=1000)].id == 1000

    def test_local_background(self):
        with pytest.warns(AstropyDeprecationWarning):
            props = source_properties(IMAGE, self.segm, localbkg_width=24)
            assert_allclose(props[0].local_background, 0., atol=1.e-7)
            props = source_properties(IMAGE << u.Jy, self.segm,
                                      localbkg_width=24)
            local_bkg = props.local_background
            assert isinstance(local_bkg, u.Quantity)
            assert_allclose(local_bkg.value, 0., atol=1.e-7)

    def test_wcs(self):
        with pytest.warns(AstropyDeprecationWarning):
            mywcs = make_wcs(IMAGE.shape)
            props = source_properties(IMAGE, self.segm, wcs=mywcs)

            assert props.sky_centroid is not None
            assert props.sky_centroid_icrs is not None
            assert props.sky_bbox_ll is not None
            assert props.sky_bbox_ul is not None
            assert props.sky_bbox_lr is not None
            assert props.sky_bbox_ur is not None

            tbl = props.to_table()
            assert len(tbl) == 3

    @pytest.mark.skipif('not HAS_GWCS')
    def test_gwcs(self):
        with pytest.warns(AstropyDeprecationWarning):
            mywcs = make_gwcs(IMAGE.shape)
            props = source_properties(IMAGE, self.segm, wcs=mywcs)
            assert props.sky_centroid is not None
            assert props.sky_centroid_icrs is not None
            assert props.sky_bbox_ll is not None
            assert props.sky_bbox_ul is not None
            assert props.sky_bbox_lr is not None
            assert props.sky_bbox_ur is not None

            tbl = props.to_table()
            assert len(tbl) == 3


@pytest.mark.skipif('not HAS_SCIPY')
class TestLegacySourceCatalog:
    def setup_class(self):
        self.segm = detect_sources(IMAGE, THRESHOLD, npixels=5)

    def test_basic(self):
        with pytest.warns(AstropyDeprecationWarning):
            segm = np.zeros(IMAGE.shape)
            x = y = np.arange(0, 100, 10)
            segm[y, x] = np.arange(10)
            cat = source_properties(IMAGE, segm)
            assert len(cat) == 9
            cat2 = cat[0:5]
            assert len(cat2) == 5
            cat3 = LegacySourceCatalog(cat2)
            del cat3[4]
            assert len(cat3) == 4

            # test iteration
            for obj in cat:
                assert obj.area.value == 1

    def test_inputs(self):
        with pytest.warns(AstropyDeprecationWarning):
            cat = source_properties(IMAGE, self.segm)
            cat2 = LegacySourceCatalog(cat[0])
            assert len(cat) == 3
            assert len(cat2) == 1

            with pytest.raises(ValueError):
                LegacySourceCatalog([])
            with pytest.raises(ValueError):
                LegacySourceCatalog('a')

    def test_table(self):
        with pytest.warns(AstropyDeprecationWarning):
            cat = source_properties(IMAGE, self.segm)
            t = cat.to_table()
            assert isinstance(t, QTable)
            assert len(t) == 3

    def test_table_include(self):
        with pytest.warns(AstropyDeprecationWarning):
            cat = source_properties(IMAGE, self.segm)
            columns = ['id', 'xcentroid']
            t = cat.to_table(columns=columns)
            assert isinstance(t, QTable)
            assert len(t) == 3
            assert t.colnames == columns

    def test_table_include_invalidname(self):
        with pytest.warns(AstropyDeprecationWarning):
            cat = source_properties(IMAGE, self.segm)
            columns = ['idzz', 'xcentroidzz']
            with pytest.raises(AttributeError):
                cat.to_table(columns=columns)

    def test_table_exclude(self):
        with pytest.warns(AstropyDeprecationWarning):
            cat = source_properties(IMAGE, self.segm)
            exclude = ['id', 'xcentroid']
            t = cat.to_table(exclude_columns=exclude)
            assert isinstance(t, QTable)
            assert len(t) == 3
            with pytest.raises(KeyError):
                t['id']

    def test_table_wcs(self):
        with pytest.warns(AstropyDeprecationWarning):
            mywcs = WCS.WCS(naxis=2)
            rho = np.pi / 3.
            scale = 0.1 / 3600.
            mywcs.wcs.cd = [[scale*np.cos(rho), -scale*np.sin(rho)],
                            [scale*np.sin(rho), scale*np.cos(rho)]]
            mywcs.wcs.ctype = ['RA---TAN', 'DEC--TAN']

            cat = source_properties(IMAGE, self.segm, wcs=mywcs)
            columns = ['sky_centroid', 'sky_centroid_icrs', 'sky_bbox_ll',
                       'sky_bbox_ul', 'sky_bbox_lr', 'sky_bbox_ur']
            t = cat.to_table(columns=columns)
            for column in columns:
                assert t[0][column] is not None
            assert t.colnames == columns

            obj = cat[0]
            row = t[0]
            assert_quantity_allclose(obj.sky_bbox_ll.ra, row['sky_bbox_ll'].ra)
            assert_quantity_allclose(obj.sky_bbox_ll.dec,
                                     row['sky_bbox_ll'].dec)

            assert_quantity_allclose(obj.sky_bbox_ul.ra, row['sky_bbox_ul'].ra)
            assert_quantity_allclose(obj.sky_bbox_ul.dec,
                                     row['sky_bbox_ul'].dec)

            assert_quantity_allclose(obj.sky_bbox_lr.ra, row['sky_bbox_lr'].ra)
            assert_quantity_allclose(obj.sky_bbox_lr.dec,
                                     row['sky_bbox_lr'].dec)

            assert_quantity_allclose(obj.sky_bbox_ur.ra, row['sky_bbox_ur'].ra)
            assert_quantity_allclose(obj.sky_bbox_ur.dec,
                                     row['sky_bbox_ur'].dec)

    def test_table_no_wcs(self):
        with pytest.warns(AstropyDeprecationWarning):
            cat = source_properties(IMAGE, self.segm)
            columns = ['sky_centroid', 'sky_centroid_icrs', 'sky_bbox_ll',
                       'sky_bbox_ul', 'sky_bbox_lr', 'sky_bbox_ur']
            t = cat.to_table(columns=columns)
            for column in columns:
                assert t[0][column] is None
            assert t.colnames == columns

    def test_repr_str(self):
        with pytest.warns(AstropyDeprecationWarning):
            segm = np.zeros(IMAGE.shape)
            x = y = np.arange(0, 100, 10)
            segm[y, x] = np.arange(10)
            cat = source_properties(IMAGE, segm)

            assert repr(cat) == str(cat)
            assert 'Catalog length:' in repr(cat)

    def test_kron_radius_all_masked(self):
        with pytest.warns(AstropyDeprecationWarning):
            mask = np.ones(IMAGE.shape, dtype=bool)
            cat = source_properties(IMAGE, self.segm, mask=mask)
            assert np.all(np.isnan(cat.kron_radius))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _utils module.
"""

import numpy as np
from numpy.testing import assert_allclose

from .._utils import mask_to_mirrored_value


def testmask_to_mirrored_value():
    center = (2.0, 2.0)
    data = np.arange(25).reshape(5, 5)
    mask = np.zeros(data.shape, dtype=bool)
    mask[0, 0] = True
    mask[1, 1] = True
    data_ref = data.copy()
    data_ref[0, 0] = data[4, 4]
    data_ref[1, 1] = data[3, 3]
    mirror_data = mask_to_mirrored_value(data, mask, center)
    assert_allclose(mirror_data, data_ref, rtol=0, atol=1.e-6)


def testmask_to_mirrored_value_range():
    """
    Test mask_to_mirrored_value when mirrored pixels are outside of the
    image.
    """
    center = (3.0, 3.0)
    data = np.arange(25).reshape(5, 5)
    mask = np.zeros(data.shape, dtype=bool)
    mask[0, 0] = True
    mask[1, 1] = True
    mask[2, 2] = True
    data_ref = data.copy()
    data_ref[0, 0] = 0.
    data_ref[1, 1] = 0.
    data_ref[2, 2] = data[4, 4]
    mirror_data = mask_to_mirrored_value(data, mask, center)
    assert_allclose(mirror_data, data_ref, rtol=0, atol=1.e-6)


def testmask_to_mirrored_value_masked():
    """
    Test mask_to_mirrored_value when mirrored pixels are also in the
    replace_mask.
    """
    center = (2.0, 2.0)
    data = np.arange(25).reshape(5, 5)
    mask = np.zeros(data.shape, dtype=bool)
    mask[0, 0] = True
    mask[1, 1] = True
    mask[3, 3] = True
    mask[4, 4] = True
    data_ref = data.copy()
    data_ref[0, 0] = 0.
    data_ref[1, 1] = 0.
    data_ref[3, 3] = 0.
    data_ref[4, 4] = 0.
    mirror_data = mask_to_mirrored_value(data, mask, center)
    mirror_data = mask_to_mirrored_value(data, mask, center)
    assert_allclose(mirror_data, data_ref, rtol=0, atol=1.e-6)


def testmask_to_mirrored_value_mask_keyword():
    """
    Test mask_to_mirrored_value when mirrored pixels are masked (via the
    mask keyword).
    """
    center = (2.0, 2.0)
    data = np.arange(25.).reshape(5, 5)
    replace_mask = np.zeros(data.shape, dtype=bool)
    mask = np.zeros(data.shape, dtype=bool)
    replace_mask[0, 2] = True
    data[4, 2] = np.nan
    mask[4, 2] = True
    result = mask_to_mirrored_value(data, replace_mask, center, mask=mask)
    assert result[0, 2] == 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides tools for convolving images with a kernel.
"""

import warnings

from astropy.convolution import Kernel2D
from astropy.units import Quantity
from astropy.utils.exceptions import AstropyUserWarning
import numpy as np


def _filter_data(data, kernel, mode='constant', fill_value=0.0,
                 check_normalization=False):
    """
    Convolve a 2D image with a 2D kernel.

    The kernel may either be a 2D `~numpy.ndarray` or a
    `~astropy.convolution.Kernel2D` object.

    Parameters
    ----------
    data : array_like
        The 2D array of the image.

    kernel : array-like (2D) or `~astropy.convolution.Kernel2D`
        The 2D kernel used to filter the input ``data``. Filtering the
        ``data`` will smooth the noise and maximize detectability of
        objects with a shape similar to the kernel.

    mode : {'constant', 'reflect', 'nearest', 'mirror', 'wrap'}, optional
        The ``mode`` determines how the array borders are handled.  For
        the ``'constant'`` mode, values outside the array borders are
        set to ``fill_value``.  The default is ``'constant'``.

    fill_value : scalar, optional
        Value to fill data values beyond the array borders if ``mode``
        is ``'constant'``.  The default is ``0.0``.

    check_normalization : bool, optional
        If `True` then a warning will be issued if the kernel is not
        normalized to 1.
    """
    from scipy import ndimage

    if kernel is not None:
        if isinstance(kernel, Kernel2D):
            kernel_array = kernel.array
        else:
            kernel_array = kernel

        if check_normalization:
            if not np.allclose(np.sum(kernel_array), 1.0):
                warnings.warn('The kernel is not normalized.',
                              AstropyUserWarning)

        # scipy.ndimage.convolve currently strips units, but be explicit
        # in case that behavior changes
        unit = None
        if isinstance(data, Quantity):
            unit = data.unit
            data = data.value

        # NOTE:  astropy.convolution.convolve fails with zero-sum
        # kernels (used in findstars) (cf. astropy #1647)
        # NOTE: if data is int and kernel is float, ndimage.convolve
        # will return an int image - here we make the data float so
        # that a float image is always returned
        result = ndimage.convolve(data.astype(float), kernel_array,
                                  mode=mode, cval=fill_value)

        if unit is not None:
            result = result * unit  # can't use *= with older astropy

        return result
    else:
        return data



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage contains tools for measuring morphological properties of
objects in an astronomical image.
"""

from .core import *  # noqa
from .non_parametric import *  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the utils module.
"""

from astropy.convolution.utils import discretize_model
from astropy.modeling.models import Gaussian2D
from astropy.table import Table
import numpy as np
from numpy.testing import assert_allclose
import pytest

from ..groupstars import DAOGroup
from ..models import IntegratedGaussianPRF
from ..photometry import BasicPSFPhotometry
from ..sandbox import DiscretePRF
from ..utils import get_grouped_psf_model, prepare_psf_model, subtract_psf
from ...utils._optional_deps import HAS_SCIPY  # noqa


PSF_SIZE = 11
GAUSSIAN_WIDTH = 1.
IMAGE_SIZE = 101

# Position and FLUXES of test sources
INTAB = Table([[50., 23, 12, 86], [50., 83, 80, 84],
               [np.pi * 10, 3.654, 20., 80 / np.sqrt(3)]],
              names=['x_0', 'y_0', 'flux_0'])

# Create test psf
psf_model = Gaussian2D(1. / (2 * np.pi * GAUSSIAN_WIDTH ** 2), PSF_SIZE // 2,
                       PSF_SIZE // 2, GAUSSIAN_WIDTH, GAUSSIAN_WIDTH)
test_psf = discretize_model(psf_model, (0, PSF_SIZE), (0, PSF_SIZE),
                            mode='oversample')

# Set up grid for test image
image = np.zeros((IMAGE_SIZE, IMAGE_SIZE))

# Add sources to test image
for x, y, flux in INTAB:
    model = Gaussian2D(flux / (2 * np.pi * GAUSSIAN_WIDTH ** 2),
                       x, y, GAUSSIAN_WIDTH, GAUSSIAN_WIDTH)
    image += discretize_model(model, (0, IMAGE_SIZE), (0, IMAGE_SIZE),
                              mode='oversample')


@pytest.fixture(scope="module")
def moffimg():
    """
    This fixture requires scipy so don't call it from non-scipy tests
    """
    from scipy import integrate
    from astropy.modeling.models import Moffat2D

    mof = Moffat2D(alpha=4.8)

    # this is the analytic value needed to get a total flux of 1
    mof.amplitude = (mof.alpha-1)/(np.pi*mof.gamma**2)

    # first make sure it really is normalized
    assert (1 - integrate.dblquad(mof, -10, 10,
                                  lambda x: -10, lambda x: 10)[0]) < 1e-6

    # now create an "image" of the PSF
    xg, yg = np.meshgrid(*([np.linspace(-2, 2, 100)]*2))

    return mof, (xg, yg, mof(xg, yg))


@pytest.mark.skipif('not HAS_SCIPY')
def test_moffat_fitting(moffimg):
    """
    Test that the Moffat to be fit in test_psf_adapter is behaving correctly
    """
    from astropy.modeling.fitting import LevMarLSQFitter
    from astropy.modeling.models import Moffat2D

    mof, (xg, yg, img) = moffimg

    # a closeish-but-wrong "guessed Moffat"
    guess_moffat = Moffat2D(x_0=.1, y_0=-.05, gamma=1.05,
                            amplitude=mof.amplitude*1.06, alpha=4.75)

    f = LevMarLSQFitter()

    fit_mof = f(guess_moffat, xg, yg, img)
    assert_allclose(fit_mof.parameters, mof.parameters, rtol=.01, atol=.0005)


# we set the tolerances in flux to be 2-3% because the shape paraameters of
# the guessed version are known to be wrong.
@pytest.mark.parametrize("prepkwargs,tols", [
                         (dict(xname='x_0', yname='y_0', fluxname=None,
                               renormalize_psf=True), (1e-3, .02)),
                         (dict(xname=None, yname=None, fluxname=None,
                               renormalize_psf=True), (1e-3, .02)),
                         (dict(xname=None, yname=None, fluxname=None,
                               renormalize_psf=False), (1e-3, .03)),
                         (dict(xname='x_0', yname='y_0', fluxname='amplitude',
                               renormalize_psf=False), (1e-3, None)),
                         ])
@pytest.mark.skipif('not HAS_SCIPY')
def test_prepare_psf_model(moffimg, prepkwargs, tols):
    """
    Test that prepare_psf_model behaves as expected for fitting (don't worry
    about full-on psf photometry for now)
    """

    from astropy.modeling.fitting import LevMarLSQFitter
    from astropy.modeling.models import Moffat2D

    mof, (xg, yg, img) = moffimg
    f = LevMarLSQFitter()

    # a close-but-wrong "guessed Moffat"
    guess_moffat = Moffat2D(x_0=.1, y_0=-.05, gamma=1.01,
                            amplitude=mof.amplitude*1.01, alpha=4.79)
    if prepkwargs['renormalize_psf']:
        # definitely very wrong, so this ensures the re-normalization
        # stuff works
        guess_moffat.amplitude = 5.

    if prepkwargs['xname'] is None:
        guess_moffat.x_0 = 0
    if prepkwargs['yname'] is None:
        guess_moffat.y_0 = 0

    psfmod = prepare_psf_model(guess_moffat, **prepkwargs)
    xytol, fluxtol = tols

    fit_psfmod = f(psfmod, xg, yg, img)

    if xytol is not None:
        assert np.abs(getattr(fit_psfmod, fit_psfmod.xname)) < xytol
        assert np.abs(getattr(fit_psfmod, fit_psfmod.yname)) < xytol
    if fluxtol is not None:
        assert np.abs(1 - getattr(fit_psfmod, fit_psfmod.fluxname)) < fluxtol

    # ensure the amplitude and shape parameters did *not* change
    assert fit_psfmod.psfmodel.gamma == guess_moffat.gamma
    assert fit_psfmod.psfmodel.alpha == guess_moffat.alpha
    if prepkwargs['fluxname'] is None:
        assert fit_psfmod.psfmodel.amplitude == guess_moffat.amplitude


@pytest.mark.skipif('not HAS_SCIPY')
def test_prepare_psf_model_offset():
    """
    Regression test to ensure the offset is in the correct direction.
    """
    norm = False
    sigma = 3.0
    amplitude = 1. / (2 * np.pi * sigma ** 2)
    xcen = ycen = 0.
    psf0 = Gaussian2D(amplitude, xcen, ycen, sigma, sigma)
    psf1 = prepare_psf_model(psf0, xname='x_mean', yname='y_mean',
                             renormalize_psf=norm)
    psf2 = prepare_psf_model(psf0, renormalize_psf=norm)
    psf3 = prepare_psf_model(psf0, xname='x_mean', renormalize_psf=norm)
    psf4 = prepare_psf_model(psf0, yname='y_mean', renormalize_psf=norm)

    yy, xx = np.mgrid[0:101, 0:101]
    psf = psf1.copy()
    xval = 48
    yval = 52
    flux = 14.51
    psf.x_mean_2 = xval
    psf.y_mean_2 = yval
    data = psf(xx, yy) * flux

    group_maker = DAOGroup(2)
    bkg_estimator = None
    fitshape = 7
    init_guesses = Table([[46.1], [57.3], [7.1]],
                         names=['x_0', 'y_0', 'flux_0'])

    phot1 = BasicPSFPhotometry(group_maker=group_maker,
                               bkg_estimator=bkg_estimator, fitshape=fitshape,
                               psf_model=psf1)
    tbl1 = phot1(image=data, init_guesses=init_guesses)

    phot2 = BasicPSFPhotometry(group_maker=group_maker,
                               bkg_estimator=bkg_estimator, fitshape=fitshape,
                               psf_model=psf2)
    tbl2 = phot2(image=data, init_guesses=init_guesses)

    phot3 = BasicPSFPhotometry(group_maker=group_maker,
                               bkg_estimator=bkg_estimator, fitshape=fitshape,
                               psf_model=psf3)
    tbl3 = phot3(image=data, init_guesses=init_guesses)

    phot4 = BasicPSFPhotometry(group_maker=group_maker,
                               bkg_estimator=bkg_estimator, fitshape=fitshape,
                               psf_model=psf4)
    tbl4 = phot4(image=data, init_guesses=init_guesses)

    assert_allclose((tbl1['x_fit'][0], tbl1['y_fit'][0], tbl1['flux_fit'][0]),
                    (xval, yval, flux))
    assert_allclose((tbl2['x_fit'][0], tbl2['y_fit'][0], tbl2['flux_fit'][0]),
                    (xval, yval, flux))
    assert_allclose((tbl3['x_fit'][0], tbl3['y_fit'][0], tbl3['flux_fit'][0]),
                    (xval, yval, flux))
    assert_allclose((tbl4['x_fit'][0], tbl4['y_fit'][0], tbl4['flux_fit'][0]),
                    (xval, yval, flux))


@pytest.mark.skipif('not HAS_SCIPY')
def test_get_grouped_psf_model():
    igp = IntegratedGaussianPRF(sigma=1.2)
    tab = Table(names=['x_0', 'y_0', 'flux_0'],
                data=[[1, 2], [3, 4], [0.5, 1]])
    pars_to_set = {'x_0': 'x_0', 'y_0': 'y_0', 'flux_0': 'flux'}

    gpsf = get_grouped_psf_model(igp, tab, pars_to_set)

    assert gpsf.x_0_0 == 1
    assert gpsf.y_0_1 == 4
    assert gpsf.flux_0 == 0.5
    assert gpsf.flux_1 == 1
    assert gpsf.sigma_0 == gpsf.sigma_1 == 1.2


@pytest.fixture(params=[0, 1, 2])
def prf_model(request):
    # use this instead of pytest.mark.parameterize as we use scipy and
    # it still calls that even if not HAS_SCIPY is set...
    prfs = [IntegratedGaussianPRF(sigma=1.2),
            Gaussian2D(x_stddev=2),
            prepare_psf_model(Gaussian2D(x_stddev=2), renormalize_psf=False)]
    return prfs[request.param]


@pytest.mark.skipif('not HAS_SCIPY')
def test_get_grouped_psf_model_submodel_names(prf_model):
    """Verify that submodel tagging works"""
    tab = Table(names=['x_0', 'y_0', 'flux_0'],
                data=[[1, 2], [3, 4], [0.5, 1]])
    pars_to_set = {'x_0': 'x_0', 'y_0': 'y_0', 'flux_0': 'flux'}

    gpsf = get_grouped_psf_model(prf_model, tab, pars_to_set)
    # There should be two submodels one named 0 and one named 1
    assert len([submodel for submodel in gpsf.traverse_postorder()
                if submodel.name == 0]) == 1
    assert len([submodel for submodel in gpsf.traverse_postorder()
                if submodel.name == 1]) == 1


@pytest.mark.skipif('not HAS_SCIPY')
def test_subtract_psf():
    """Test subtract_psf."""

    prf = DiscretePRF(test_psf, subsampling=1)
    posflux = INTAB.copy()
    for n in posflux.colnames:
        posflux.rename_column(n, n.split('_')[0] + '_fit')
    residuals = subtract_psf(image, prf, posflux)
    assert_allclose(residuals, np.zeros_like(image), atol=1E-4)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage contains tools for making or loading datasets for
examples and tests.
"""

from .load import *  # noqa
from .make import *  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module implements the DAOStarFinder class.
"""
import inspect
import warnings

from astropy.nddata import extract_array
from astropy.table import QTable
from astropy.utils import lazyproperty
import numpy as np

from .core import StarFinderBase, _StarFinderKernel
from ..utils._convolution import _filter_data
from ..utils.exceptions import NoDetectionsWarning
from ..utils._misc import _get_version_info

__all__ = ['DAOStarFinder']


class DAOStarFinder(StarFinderBase):
    """
    Detect stars in an image using the DAOFIND (`Stetson 1987
    <https://ui.adsabs.harvard.edu/abs/1987PASP...99..191S/abstract>`_)
    algorithm.

    DAOFIND (`Stetson 1987; PASP 99, 191
    <https://ui.adsabs.harvard.edu/abs/1987PASP...99..191S/abstract>`_)
    searches images for local density maxima that have a peak amplitude
    greater than ``threshold`` (approximately; ``threshold`` is applied
    to a convolved image) and have a size and shape similar to the
    defined 2D Gaussian kernel.  The Gaussian kernel is defined by the
    ``fwhm``, ``ratio``, ``theta``, and ``sigma_radius`` input
    parameters.

    ``DAOStarFinder`` finds the object centroid by fitting the marginal x
    and y 1D distributions of the Gaussian kernel to the marginal x and
    y distributions of the input (unconvolved) ``data`` image.

    ``DAOStarFinder`` calculates the object roundness using two methods. The
    ``roundlo`` and ``roundhi`` bounds are applied to both measures of
    roundness.  The first method (``roundness1``; called ``SROUND`` in
    `DAOFIND`_) is based on the source symmetry and is the ratio of a
    measure of the object's bilateral (2-fold) to four-fold symmetry.
    The second roundness statistic (``roundness2``; called ``GROUND`` in
    `DAOFIND`_) measures the ratio of the difference in the height of
    the best fitting Gaussian function in x minus the best fitting
    Gaussian function in y, divided by the average of the best fitting
    Gaussian functions in x and y.  A circular source will have a zero
    roundness.  A source extended in x or y will have a negative or
    positive roundness, respectively.

    The sharpness statistic measures the ratio of the difference between
    the height of the central pixel and the mean of the surrounding
    non-bad pixels in the convolved image, to the height of the best
    fitting Gaussian function at that point.

    Parameters
    ----------
    threshold : float
        The absolute image value above which to select sources.

    fwhm : float
        The full-width half-maximum (FWHM) of the major axis of the
        Gaussian kernel in units of pixels.

    ratio : float, optional
        The ratio of the minor to major axis standard deviations of the
        Gaussian kernel.  ``ratio`` must be strictly positive and less
        than or equal to 1.0.  The default is 1.0 (i.e., a circular
        Gaussian kernel).

    theta : float, optional
        The position angle (in degrees) of the major axis of the
        Gaussian kernel measured counter-clockwise from the positive x
        axis.

    sigma_radius : float, optional
        The truncation radius of the Gaussian kernel in units of sigma
        (standard deviation) [``1 sigma = FWHM /
        (2.0*sqrt(2.0*log(2.0)))``].

    sharplo : float, optional
        The lower bound on sharpness for object detection.

    sharphi : float, optional
        The upper bound on sharpness for object detection.

    roundlo : float, optional
        The lower bound on roundness for object detection.

    roundhi : float, optional
        The upper bound on roundness for object detection.

    sky : float, optional
        The background sky level of the image.  Setting ``sky`` affects
        only the output values of the object ``peak``, ``flux``, and
        ``mag`` values.  The default is 0.0, which should be used to
        replicate the results from `DAOFIND`_.

    exclude_border : bool, optional
        Set to `True` to exclude sources found within half the size of
        the convolution kernel from the image borders.  The default is
        `False`, which is the mode used by `DAOFIND`_.

    brightest : int, None, optional
        Number of brightest objects to keep after sorting the full object list.
        If ``brightest`` is set to `None`, all objects will be selected.

    peakmax : float, None, optional
        Maximum peak pixel value in an object. Only objects whose peak pixel
        values are *strictly smaller* than ``peakmax`` will be selected.
        This may be used to exclude saturated sources. By default, when
        ``peakmax`` is set to `None`, all objects will be selected.

        .. warning::
            `DAOStarFinder` automatically excludes objects whose peak
            pixel values are negative. Therefore, setting ``peakmax`` to a
            non-positive value would result in exclusion of all objects.

    xycoords : `None` or Nx2 `~numpy.ndarray`
        The (x, y) pixel coordinates of the approximate centroid
        positions of identified sources. If ``xycoords`` are input, the
        algorithm will skip the source-finding step.

    See Also
    --------
    IRAFStarFinder

    Notes
    -----
    For the convolution step, this routine sets pixels beyond the image
    borders to 0.0.  The equivalent parameters in `DAOFIND`_ are
    ``boundary='constant'`` and ``constant=0.0``.

    The main differences between `~photutils.detection.DAOStarFinder`
    and `~photutils.detection.IRAFStarFinder` are:

    * `~photutils.detection.IRAFStarFinder` always uses a 2D
      circular Gaussian kernel, while
      `~photutils.detection.DAOStarFinder` can use an elliptical
      Gaussian kernel.

    * `~photutils.detection.IRAFStarFinder` calculates the objects'
      centroid, roundness, and sharpness using image moments.

    References
    ----------
    .. [1] Stetson, P. 1987; PASP 99, 191
           (https://ui.adsabs.harvard.edu/abs/1987PASP...99..191S/abstract)
    .. [2] https://iraf.net/irafhelp.php?val=daofind

    .. _DAOFIND: https://iraf.net/irafhelp.php?val=daofind
    """

    def __init__(self, threshold, fwhm, ratio=1.0, theta=0.0,
                 sigma_radius=1.5, sharplo=0.2, sharphi=1.0, roundlo=-1.0,
                 roundhi=1.0, sky=0.0, exclude_border=False,
                 brightest=None, peakmax=None, xycoords=None):

        if not np.isscalar(threshold):
            raise TypeError('threshold must be a scalar value.')

        if not np.isscalar(fwhm):
            raise TypeError('fwhm must be a scalar value.')

        self.threshold = threshold
        self.fwhm = fwhm
        self.ratio = ratio
        self.theta = theta
        self.sigma_radius = sigma_radius
        self.sharplo = sharplo
        self.sharphi = sharphi
        self.roundlo = roundlo
        self.roundhi = roundhi
        self.sky = sky
        self.exclude_border = exclude_border
        self.brightest = self._validate_brightest(brightest)
        self.peakmax = peakmax

        if xycoords is not None:
            xycoords = np.asarray(xycoords)
            if xycoords.ndim != 2 or xycoords.shape[1] != 2:
                raise ValueError('xycoords must be shaped as a Nx2 array')
        self.xycoords = xycoords

        self.kernel = _StarFinderKernel(self.fwhm, self.ratio, self.theta,
                                        self.sigma_radius)
        self.threshold_eff = self.threshold * self.kernel.relerr

    @staticmethod
    def _validate_brightest(brightest):
        if brightest is not None:
            if brightest <= 0:
                raise ValueError('brightest must be >= 0')
            bright_int = int(brightest)
            if bright_int != brightest:
                raise ValueError('brightest must be an integer')
            brightest = bright_int
        return brightest

    def _get_raw_catalog(self, data, mask=None):
        convolved_data = _filter_data(data, self.kernel.data, mode='constant',
                                      fill_value=0.0,
                                      check_normalization=False)

        if self.xycoords is None:
            xypos = self._find_stars(convolved_data, self.kernel,
                                     self.threshold_eff, mask=mask,
                                     exclude_border=self.exclude_border)
        else:
            xypos = self.xycoords

        if xypos is None:
            warnings.warn('No sources were found.', NoDetectionsWarning)
            return None

        cat = _DAOStarFinderCatalog(data, convolved_data, xypos, self.kernel,
                                    self.threshold, sky=self.sky,
                                    sharplo=self.sharplo, sharphi=self.sharphi,
                                    roundlo=self.roundlo, roundhi=self.roundhi,
                                    brightest=self.brightest,
                                    peakmax=self.peakmax)
        return cat

    def find_stars(self, data, mask=None):
        """
        Find stars in an astronomical image.

        Parameters
        ----------
        data : 2D array_like
            The 2D image array.

        mask : 2D bool array, optional
            A boolean mask with the same shape as ``data``, where a
            `True` value indicates the corresponding element of ``data``
            is masked.  Masked pixels are ignored when searching for
            stars.

        Returns
        -------
        table : `~astropy.table.QTable` or `None`
            A table of found stars with the following parameters:

            * ``id``: unique object identification number.
            * ``xcentroid, ycentroid``: object centroid.
            * ``sharpness``: object sharpness.
            * ``roundness1``: object roundness based on symmetry.
            * ``roundness2``: object roundness based on marginal Gaussian
              fits.
            * ``npix``: the total number of pixels in the Gaussian kernel
              array.
            * ``sky``: the input ``sky`` parameter.
            * ``peak``: the peak, sky-subtracted, pixel value of the object.
            * ``flux``: the object flux calculated as the peak density in
              the convolved image divided by the detection threshold.  This
              derivation matches that of `DAOFIND`_ if ``sky`` is 0.0.
            * ``mag``: the object instrumental magnitude calculated as
              ``-2.5 * log10(flux)``.  The derivation matches that of
              `DAOFIND`_ if ``sky`` is 0.0.

            `None` is returned if no stars are found.
        """
        cat = self._get_raw_catalog(data, mask=mask)
        if cat is None:
            return None

        # apply all selection filters
        cat = cat.apply_all_filters()
        if cat is None:
            return None

        # create the output table
        return cat.to_table()


class _DAOStarFinderCatalog:
    """
    Class to create a catalog of the properties of each detected star,
    as defined by `DAOFIND`_.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D image.

    convolved_data : 2D `~numpy.ndarray`
        The convolved 2D image.

    xypos: Nx2 `numpy.ndarray`
        A Nx2 array of (x, y) pixel coordinates denoting the central
        positions of the stars.

    kernel : `_StarFinderKernel`
        The convolution kernel. This kernel must match the kernel used
        to create the ``convolved_data``.

    threshold : float
        The absolute image value above which sources were selected.

    sky : float, optional
        The local sky level around the source.  ``sky`` is used only to
        calculate the source peak value, flux, and magnitude.  The
        default is 0.

    .. _DAOFIND: https://iraf.net/irafhelp.php?val=daofind
    """

    def __init__(self, data, convolved_data, xypos, kernel, threshold,
                 sky=0., sharplo=0.2, sharphi=1.0, roundlo=-1.0, roundhi=1.0,
                 brightest=None, peakmax=None):

        self.data = data
        self.convolved_data = convolved_data
        self.xypos = np.atleast_2d(xypos)
        self.kernel = kernel
        self.threshold = threshold
        self._sky = sky  # DAOFIND has no sky input -> same as sky=0.
        self.sharplo = sharplo
        self.sharphi = sharphi
        self.roundlo = roundlo
        self.roundhi = roundhi
        self.brightest = brightest
        self.peakmax = peakmax

        self.id = np.arange(len(self)) + 1
        self.threshold_eff = threshold * kernel.relerr
        self.cutout_shape = kernel.shape
        self.cutout_center = tuple([(size - 1) // 2 for size in kernel.shape])
        self.default_columns = ('id', 'xcentroid', 'ycentroid', 'sharpness',
                                'roundness1', 'roundness2', 'npix', 'sky',
                                'peak', 'flux', 'mag')

    def __len__(self):
        return len(self.xypos)

    def __getitem__(self, index):
        newcls = object.__new__(self.__class__)
        init_attr = ('data', 'convolved_data', 'kernel', 'threshold', '_sky',
                     'sharplo', 'sharphi', 'roundlo', 'roundhi', 'brightest',
                     'peakmax', 'threshold_eff', 'cutout_shape',
                     'cutout_center', 'default_columns')
        for attr in init_attr:
            setattr(newcls, attr, getattr(self, attr))

        # xypos determines ordering and isscalar
        # NOTE: always keep as a 2D array, even for a single source
        attr = 'xypos'
        value = getattr(self, attr)[index]
        setattr(newcls, attr, np.atleast_2d(value))

        keys = set(self.__dict__.keys()) & set(self._lazyproperties)
        keys.add('id')
        for key in keys:
            value = self.__dict__[key]

            # do not insert lazy attributes that are always scalar (e.g.,
            # isscalar), i.e., not an array/list for each source
            if np.isscalar(value):
                continue

            # value is always at least a 1D array, even for a single source
            value = np.atleast_1d(value[index])

            newcls.__dict__[key] = value
        return newcls

    @lazyproperty
    def isscalar(self):
        """
        Whether the instance is scalar (e.g., a single source).
        """
        return self.xypos.shape == (1, 2)

    @property
    def _lazyproperties(self):
        """
        Return all lazyproperties (even in superclasses).
        """
        def islazyproperty(obj):
            return isinstance(obj, lazyproperty)
        return [i[0] for i in inspect.getmembers(self.__class__,
                                                 predicate=islazyproperty)]

    def reset_ids(self):
        """Reset the ID column to be consecutive integers."""
        self.id = np.arange(len(self)) + 1

    def make_cutouts(self, data):
        cutouts = []
        for xpos, ypos in self.xypos:
            cutouts.append(extract_array(data, self.cutout_shape, (ypos, xpos),
                                         fill_value=0.0))
        return np.array(cutouts)

    @lazyproperty
    def cutout_data(self):
        return self.make_cutouts(self.data)

    @lazyproperty
    def cutout_convdata(self):
        return self.make_cutouts(self.convolved_data)

    @lazyproperty
    def data_peak(self):
        return self.cutout_data[:, self.cutout_center[0],
                                self.cutout_center[1]]

    @lazyproperty
    def convdata_peak(self):
        return self.cutout_convdata[:, self.cutout_center[0],
                                    self.cutout_center[1]]

    @lazyproperty
    def roundness1(self):
        # set the central (peak) pixel to zero for the sum4 calculation
        cutout_conv = self.cutout_convdata.copy()
        cutout_conv[:, self.cutout_center[0], self.cutout_center[1]] = 0.0

        # calculate the four roundness quadrants.
        # the cutout size always matches the kernel size, which has odd
        # dimensions.
        # quad1 = bottom right
        # quad2 = bottom left
        # quad3 = top left
        # quad4 = top right
        # 3 3 4 4 4
        # 3 3 4 4 4
        # 3 3 x 1 1
        # 2 2 2 1 1
        # 2 2 2 1 1
        quad1 = cutout_conv[:, 0:self.cutout_center[0] + 1,
                            self.cutout_center[1] + 1:]
        quad2 = cutout_conv[:, 0:self.cutout_center[0],
                            0:self.cutout_center[1] + 1]
        quad3 = cutout_conv[:, self.cutout_center[0]:,
                            0:self.cutout_center[1]]
        quad4 = cutout_conv[:, self.cutout_center[0] + 1:,
                            self.cutout_center[1]:]

        axis = (1, 2)
        sum2 = (-quad1.sum(axis=axis) + quad2.sum(axis=axis)
                - quad3.sum(axis=axis) + quad4.sum(axis=axis))
        sum4 = np.abs(cutout_conv).sum(axis=axis)

        # ignore divide-by-zero RuntimeWarning
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            roundness1 = 2.0 * sum2 / sum4

        return roundness1

    @lazyproperty
    def sharpness(self):
        # mean value of the unconvolved data (excluding the peak)
        cutout_data_masked = self.cutout_data * self.kernel.mask
        data_mean = ((np.sum(cutout_data_masked, axis=(1, 2)) - self.data_peak)
                     / (self.kernel.npixels - 1))

        return (self.data_peak - data_mean) / self.convdata_peak

    def daofind_marginal_fit(self, axis=0):
        """
        Fit 1D Gaussians, defined from the marginal x/y kernel
        distributions, to the marginal x/y distributions of the original
        (unconvolved) image.

        These fits are used calculate the star centroid and roundness2
        ("GROUND") properties.

        Parameters
        ----------
        axis : {0, 1}, optional
            The axis for which the marginal fit is performed:

            * 0: for the x axis
            * 1: for the y axis

        Returns
        -------
        dx : float
            The fractional shift in x or y (depending on ``axis`` value)
            of the image centroid relative to the maximum pixel.

        hx : float
            The height of the best-fitting Gaussian to the marginal x or
            y (depending on ``axis`` value) distribution of the
            unconvolved source data.
        """
        # define triangular weighting functions along each axis, peaked
        # in the middle and equal to one at the edge
        ycen, xcen = self.cutout_center
        xx = xcen - np.abs(np.arange(self.cutout_shape[1]) - xcen) + 1
        yy = ycen - np.abs(np.arange(self.cutout_shape[0]) - ycen) + 1
        xwt, ywt = np.meshgrid(xx, yy)

        if axis == 0:  # marginal distributions along x axis
            wt = xwt[0]  # 1D
            wts = ywt  # 2D
            size = self.cutout_shape[1]
            center = xcen
            sigma = self.kernel.xsigma
            dxx = center - np.arange(size)
        elif axis == 1:  # marginal distributions along y axis
            wt = np.transpose(ywt)[0]  # 1D
            wts = xwt  # 2D
            size = self.cutout_shape[0]
            center = ycen
            sigma = self.kernel.ysigma
            dxx = np.arange(size) - center

        # compute marginal sums for given axis
        wt_sum = np.sum(wt)
        dx = center - np.arange(size)

        # weighted marginal sums
        kern_sum_1d = np.sum(self.kernel.gaussian_kernel_unmasked * wts,
                             axis=axis)
        kern_sum = np.sum(kern_sum_1d * wt)
        kern2_sum = np.sum(kern_sum_1d**2 * wt)

        dkern_dx = kern_sum_1d * dx
        dkern_dx_sum = np.sum(dkern_dx * wt)
        dkern_dx2_sum = np.sum(dkern_dx**2 * wt)
        kern_dkern_dx_sum = np.sum(kern_sum_1d * dkern_dx * wt)

        data_sum_1d = np.sum(self.cutout_data * wts, axis=axis + 1)
        data_sum = np.sum(data_sum_1d * wt, axis=1)
        data_kern_sum = np.sum(data_sum_1d * kern_sum_1d * wt, axis=1)
        data_dkern_dx_sum = np.sum(data_sum_1d * dkern_dx * wt, axis=1)
        data_dx_sum = np.sum(data_sum_1d * dxx * wt, axis=1)

        # perform linear least-squares fit (where data = sky + hx*kernel)
        # to find the amplitude (hx)
        hx_numer = data_kern_sum - (data_sum * kern_sum) / wt_sum
        hx_denom = kern2_sum - (kern_sum**2 / wt_sum)

        # reject the star if the fit amplitude is not positive
        mask1 = (hx_numer <= 0.) | (hx_denom <= 0.)

        # ignore divide-by-zero RuntimeWarning
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            # compute fit amplitude
            hx = hx_numer / hx_denom

            # sky = (data_sum - (hx * kern_sum)) / wt_sum

            # compute centroid shift
            dx = ((kern_dkern_dx_sum
                   - (data_dkern_dx_sum - dkern_dx_sum * data_sum))
                  / (hx * dkern_dx2_sum / sigma**2))

            dx2 = data_dx_sum / data_sum

        hsize = size / 2.0
        mask2 = (np.abs(dx) > hsize)
        mask3 = (data_sum == 0.)
        mask4 = (mask2 & mask3)
        mask5 = (mask2 & ~mask3)

        dx[mask4] = 0.0
        dx[mask5] = dx2[mask5]
        mask6 = (np.abs(dx) > hsize)
        dx[mask6] = 0.0

        hx[mask1] = np.nan
        dx[mask1] = np.nan

        return np.transpose((dx, hx))

    @lazyproperty
    def dx_hx(self):
        return self.daofind_marginal_fit(axis=0)

    @lazyproperty
    def dy_hy(self):
        return self.daofind_marginal_fit(axis=1)

    @lazyproperty
    def dx(self):
        return np.transpose(self.dx_hx)[0]

    @lazyproperty
    def dy(self):
        return np.transpose(self.dy_hy)[0]

    @lazyproperty
    def hx(self):
        return np.transpose(self.dx_hx)[1]

    @lazyproperty
    def hy(self):
        return np.transpose(self.dy_hy)[1]

    @lazyproperty
    def xcentroid(self):
        return np.transpose(self.xypos)[0] + self.dx

    @lazyproperty
    def ycentroid(self):
        return np.transpose(self.xypos)[1] + self.dy

    @lazyproperty
    def roundness2(self):
        """
        The star roundness.

        This roundness parameter represents the ratio of the difference
        in the height of the best fitting Gaussian function in x minus
        the best fitting Gaussian function in y, divided by the average
        of the best fitting Gaussian functions in x and y.  A circular
        source will have a zero roundness.  A source extended in x or y
        will have a negative or positive roundness, respectively.
        """
        return 2.0 * (self.hx - self.hy) / (self.hx + self.hy)

    @lazyproperty
    def peak(self):
        return self.data_peak - self.sky

    @lazyproperty
    def flux(self):
        return ((self.convdata_peak / self.threshold_eff)
                - (self.sky * self.npix))

    @lazyproperty
    def mag(self):
        # ignore RunTimeWarning if flux is <= 0
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=RuntimeWarning)
            mag = -2.5 * np.log10(self.flux)
            mag[self.flux <= 0] = np.nan
        return mag

    @lazyproperty
    def sky(self):
        return np.full(len(self), fill_value=self._sky)

    @lazyproperty
    def npix(self):
        return np.full(len(self), fill_value=self.kernel.data.size)

    def apply_filters(self):
        """Filter the catalog."""
        mask = (~np.isnan(self.dx) & ~np.isnan(self.dy)
                & ~np.isnan(self.hx) & ~np.isnan(self.hy))
        mask &= ((self.sharpness > self.sharplo)
                 & (self.sharpness < self.sharphi)
                 & (self.roundness1 > self.roundlo)
                 & (self.roundness1 < self.roundhi)
                 & (self.roundness2 > self.roundlo)
                 & (self.roundness2 < self.roundhi))
        if self.peakmax is not None:
            mask &= (self.peak < self.peakmax)
        newcat = self[mask]

        if len(newcat) == 0:
            warnings.warn('Sources were found, but none pass the sharpness, '
                          'roundness, or peakmax criteria',
                          NoDetectionsWarning)
            return None

        return newcat

    def select_brightest(self):
        """
        Sort the catalog by the brightest fluxes and select the
        top brightest sources.
        """
        newcat = self
        if self.brightest is not None:
            idx = np.argsort(self.flux)[::-1][:self.brightest]
            newcat = self[idx]
        return newcat

    def apply_all_filters(self):
        """
        Apply all filters, select the brightest, and reset the source
        ids.
        """
        cat = self.apply_filters()
        if cat is None:
            return None
        cat = cat.select_brightest()
        cat.reset_ids()
        return cat

    def to_table(self, columns=None):
        meta = {'version': _get_version_info()}
        table = QTable(meta=meta)
        if columns is None:
            columns = self.default_columns
        for column in columns:
            table[column] = getattr(self, column)
        return table



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the detect module.
"""

from astropy.convolution import Gaussian2DKernel
from astropy.stats import gaussian_fwhm_to_sigma
from astropy.tests.helper import catch_warnings
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import pytest

from ...utils.exceptions import NoDetectionsWarning
from ..detect import detect_threshold, detect_sources, make_source_mask
from ...datasets import make_4gaussians_image
from ...utils._optional_deps import HAS_SCIPY  # noqa


DATA = np.array([[0, 1, 0], [0, 2, 0], [0, 0, 0]]).astype(float)
REF1 = np.array([[0, 0, 0], [0, 1, 0], [0, 0, 0]])


@pytest.mark.skipif('not HAS_SCIPY')
class TestDetectThreshold:
    def test_nsigma(self):
        """Test basic nsigma."""

        threshold = detect_threshold(DATA, nsigma=0.1)
        ref = 0.4 * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_nsigma_zero(self):
        """Test nsigma=0."""

        threshold = detect_threshold(DATA, nsigma=0.0)
        ref = (1. / 3.) * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_background(self):
        threshold = detect_threshold(DATA, nsigma=1.0, background=1)
        ref = (5. / 3.) * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_background_image(self):
        background = np.ones((3, 3))
        threshold = detect_threshold(DATA, nsigma=1.0, background=background)
        ref = (5. / 3.) * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_background_badshape(self):
        wrong_shape = np.zeros((2, 2))
        with pytest.raises(ValueError):
            detect_threshold(DATA, nsigma=2., background=wrong_shape)

    def test_error(self):
        threshold = detect_threshold(DATA, nsigma=1.0, error=1)
        ref = (4. / 3.) * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_error_image(self):
        error = np.ones((3, 3))
        threshold = detect_threshold(DATA, nsigma=1.0, error=error)
        ref = (4. / 3.) * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_error_badshape(self):
        wrong_shape = np.zeros((2, 2))
        with pytest.raises(ValueError):
            detect_threshold(DATA, nsigma=2., error=wrong_shape)

    def test_background_error(self):
        threshold = detect_threshold(DATA, nsigma=2.0, background=10.,
                                     error=1.)
        ref = 12. * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_background_error_images(self):
        background = np.ones((3, 3)) * 10.
        error = np.ones((3, 3))
        threshold = detect_threshold(DATA, nsigma=2.0, background=background,
                                     error=error)
        ref = 12. * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_mask_value(self):
        """Test detection with mask_value."""

        threshold = detect_threshold(DATA, nsigma=1.0, mask_value=0.0)
        ref = 2. * np.ones((3, 3))
        assert_array_equal(threshold, ref)

    def test_image_mask(self):
        """
        Test detection with image_mask.
        Set sigma=10 and iters=1 to prevent sigma clipping after
        applying the mask.
        """

        mask = REF1.astype(bool)
        threshold = detect_threshold(DATA, nsigma=1., error=0, mask=mask,
                                     sigclip_sigma=10, sigclip_iters=1)
        ref = (1. / 8.) * np.ones((3, 3))
        assert_array_equal(threshold, ref)

    def test_image_mask_override(self):
        """Test that image_mask overrides mask_value."""

        mask = REF1.astype(bool)
        threshold = detect_threshold(DATA, nsigma=0.1, error=0, mask_value=0.0,
                                     mask=mask, sigclip_sigma=10,
                                     sigclip_iters=1)
        ref = np.ones((3, 3))
        assert_array_equal(threshold, ref)


@pytest.mark.skipif('not HAS_SCIPY')
class TestDetectSources:
    def setup_class(self):
        self.data = np.array([[0, 1, 0], [0, 2, 0],
                              [0, 0, 0]]).astype(float)
        self.refdata = np.array([[0, 1, 0], [0, 1, 0], [0, 0, 0]])

        fwhm2sigma = 1.0 / (2.0 * np.sqrt(2.0 * np.log(2.0)))
        kernel = Gaussian2DKernel(2. * fwhm2sigma, x_size=3, y_size=3)
        kernel.normalize()
        self.kernel = kernel

    def test_detection(self):
        """Test basic detection."""

        segm = detect_sources(self.data, threshold=0.9, npixels=2)
        assert_array_equal(segm.data, self.refdata)

    def test_small_sources(self):
        """Test detection where sources are smaller than npixels size."""

        with catch_warnings(NoDetectionsWarning) as warning_lines:
            detect_sources(self.data, threshold=0.9, npixels=5)
            assert warning_lines[0].category == NoDetectionsWarning
            assert 'No sources were found.' in str(warning_lines[0].message)

    def test_npixels(self):
        """
        Test removal of sources whose size is less than npixels.
        Regression tests for #663.
        """

        data = np.zeros((8, 8))
        data[0:4, 0] = 1
        data[0, 0:4] = 1
        data[3, 3:] = 2
        data[3:, 3] = 2

        segm = detect_sources(data, 0, npixels=8)
        assert segm.nlabels == 1
        segm = detect_sources(data, 0, npixels=9)
        assert segm.nlabels == 1

        data = np.zeros((8, 8))
        data[0:4, 0] = 1
        data[0, 0:4] = 1
        data[3, 2:] = 2
        data[3:, 2] = 2
        data[5:, 3] = 2

        npixels = np.arange(9, 14)
        for npixels in np.arange(9, 14):
            segm = detect_sources(data, 0, npixels=npixels)
            assert segm.nlabels == 1
            assert segm.areas[0] == 13

        with catch_warnings(NoDetectionsWarning) as warning_lines:
            detect_sources(data, 0, npixels=14)
            assert warning_lines[0].category == NoDetectionsWarning
            assert 'No sources were found.' in str(warning_lines[0].message)

    def test_zerothresh(self):
        """Test detection with zero threshold."""

        segm = detect_sources(self.data, threshold=0., npixels=2)
        assert_array_equal(segm.data, self.refdata)

    def test_zerodet(self):
        """Test detection with large threshold giving no detections."""

        with catch_warnings(NoDetectionsWarning) as warning_lines:
            detect_sources(self.data, threshold=7, npixels=2)
            assert warning_lines[0].category == NoDetectionsWarning
            assert 'No sources were found.' in str(warning_lines[0].message)

    def test_8connectivity(self):
        """Test detection with connectivity=8."""

        data = np.eye(3)
        segm = detect_sources(data, threshold=0.9, npixels=1, connectivity=8)
        assert_array_equal(segm.data, data)

    def test_4connectivity(self):
        """Test detection with connectivity=4."""

        data = np.eye(3)
        ref = np.diag([1, 2, 3])
        segm = detect_sources(data, threshold=0.9, npixels=1, connectivity=4)
        assert_array_equal(segm.data, ref)

    def test_basic_kernel(self):
        """Test detection with kernel."""

        kernel = np.ones((3, 3)) / 9.
        threshold = 0.3
        expected = np.ones((3, 3))
        expected[2] = 0
        segm = detect_sources(self.data, threshold, npixels=1, kernel=kernel)
        assert_array_equal(segm.data, expected)

    def test_npixels_nonint(self):
        """Test if error raises if npixel is non-integer."""

        with pytest.raises(ValueError):
            detect_sources(self.data, threshold=1, npixels=0.1)

    def test_npixels_negative(self):
        """Test if error raises if npixel is negative."""

        with pytest.raises(ValueError):
            detect_sources(self.data, threshold=1, npixels=-1)

    def test_connectivity_invalid(self):
        """Test if error raises if connectivity is invalid."""

        with pytest.raises(ValueError):
            detect_sources(self.data, threshold=1, npixels=1, connectivity=10)

    def test_kernel_array(self):
        segm = detect_sources(self.data, 0.1, npixels=1,
                              kernel=self.kernel.array)
        assert_array_equal(segm.data, np.ones((3, 3)))

    def test_kernel(self):
        segm = detect_sources(self.data, 0.1, npixels=1, kernel=self.kernel)
        assert_array_equal(segm.data, np.ones((3, 3)))

    def test_mask(self):
        data = np.zeros((11, 11))
        data[3:8, 3:8] = 5.
        mask = np.zeros(data.shape, dtype=bool)
        mask[4:6, 4:6] = True
        segm1 = detect_sources(data, 1., 1.)
        segm2 = detect_sources(data, 1., 1., mask=mask)
        assert segm2.areas[0] == segm1.areas[0] - mask.sum()

    def test_mask_shape(self):
        with pytest.raises(ValueError):
            detect_sources(self.data, 1., 1., mask=np.ones((5, 5)))


@pytest.mark.skipif('not HAS_SCIPY')
class TestMakeSourceMask:
    def setup_class(self):
        self.data = make_4gaussians_image()

    def test_dilate_size(self):
        mask1 = make_source_mask(self.data, 5, 10)
        mask2 = make_source_mask(self.data, 5, 10, dilate_size=20)
        assert np.count_nonzero(mask2) > np.count_nonzero(mask1)

    def test_kernel(self):
        mask1 = make_source_mask(self.data, 5, 10, filter_fwhm=2,
                                 filter_size=3)
        sigma = 2 * gaussian_fwhm_to_sigma
        kernel = Gaussian2DKernel(sigma, x_size=3, y_size=3)
        mask2 = make_source_mask(self.data, 5, 10, kernel=kernel)
        assert_allclose(mask1, mask2)

    def test_no_detections(self):
        with catch_warnings(NoDetectionsWarning) as warning_lines:
            mask = make_source_mask(self.data, 100, 100)
            assert np.count_nonzero(mask) == 0
            assert warning_lines[0].category == NoDetectionsWarning



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the circle module.
"""

from astropy.coordinates import SkyCoord
import astropy.units as u
import numpy as np
from numpy.testing import assert_allclose
import pytest

from .test_aperture_common import BaseTestAperture
from ..circle import (CircularAperture, CircularAnnulus, SkyCircularAperture,
                      SkyCircularAnnulus)
from ...utils._optional_deps import HAS_MATPLOTLIB  # noqa


POSITIONS = [(10, 20), (30, 40), (50, 60), (70, 80)]
RA, DEC = np.transpose(POSITIONS)
SKYCOORD = SkyCoord(ra=RA, dec=DEC, unit='deg')
UNIT = u.arcsec


class TestCircularAperture(BaseTestAperture):
    aperture = CircularAperture(POSITIONS, r=3.)

    @pytest.mark.skipif('not HAS_MATPLOTLIB')
    def test_plot(self):
        self.aperture.plot()

    @pytest.mark.skipif('not HAS_MATPLOTLIB')
    def test_plot_returns_patches(self):
        from matplotlib import pyplot as plt
        from matplotlib.patches import Patch

        my_patches = self.aperture.plot()
        assert isinstance(my_patches, list)
        for patch in my_patches:
            assert isinstance(patch, Patch)

        # test creating a legend with these patches
        plt.legend(my_patches, list(range(len(my_patches))))


class TestCircularAnnulus(BaseTestAperture):
    aperture = CircularAnnulus(POSITIONS, r_in=3., r_out=7.)

    @pytest.mark.skipif('not HAS_MATPLOTLIB')
    def test_plot(self):
        self.aperture.plot()

    @pytest.mark.skipif('not HAS_MATPLOTLIB')
    def test_plot_returns_patches(self):
        from matplotlib import pyplot as plt
        from matplotlib.patches import Patch

        my_patches = self.aperture.plot()
        assert isinstance(my_patches, list)

        for p in my_patches:
            assert isinstance(p, Patch)

        # make sure I can create a legend with these patches
        labels = list(range(len(my_patches)))
        plt.legend(my_patches, labels)


class TestSkyCircularAperture(BaseTestAperture):
    aperture = SkyCircularAperture(SKYCOORD, r=3.*UNIT)


class TestSkyCircularAnnulus(BaseTestAperture):
    aperture = SkyCircularAnnulus(SKYCOORD, r_in=3.*UNIT, r_out=7.*UNIT)


def test_slicing():
    xypos = [(10, 10), (20, 20), (30, 30)]
    aper1 = CircularAperture(xypos, r=3)
    aper2 = aper1[0:2]
    assert len(aper2) == 2

    aper3 = aper1[0]
    assert aper3.isscalar
    with pytest.raises(TypeError):
        len(aper3)

    with pytest.raises(TypeError):
        _ = aper3[0]


def test_area_overlap():
    data = np.ones((11, 11))
    xypos = [(0, 0), (5, 5), (50, 50)]
    aper = CircularAperture(xypos, r=3)
    areas = aper.area_overlap(data)
    assert_allclose(areas, [10.304636, np.pi*9., np.nan])

    aper2 = CircularAperture(xypos[1], r=3)
    area2 = aper2.area_overlap(data)
    assert_allclose(area2, np.pi * 9.)


def test_area_overlap_mask():
    data = np.ones((11, 11))
    mask = np.zeros((11, 11), dtype=bool)
    mask[0, 0:2] = True
    mask[5, 5:7] = True
    xypos = [(0, 0), (5, 5), (50, 50)]
    aper = CircularAperture(xypos, r=3)
    areas = aper.area_overlap(data, mask=mask)
    areas_exp = np.array([10.304636, np.pi*9., np.nan]) - 2.
    assert_allclose(areas, areas_exp)

    with pytest.raises(ValueError):
        mask = np.zeros((3, 3), dtype=bool)
        aper.area_overlap(data, mask=mask)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the windows module.
"""

import numpy as np
from numpy.testing import assert_allclose
import pytest

from ..windows import (CosineBellWindow, HanningWindow, SplitCosineBellWindow,
                       TopHatWindow, TukeyWindow)
from ....utils._optional_deps import HAS_SCIPY  # noqa


def test_hanning():
    win = HanningWindow()
    data = win((5, 5))
    ref = [0., 0.19715007, 0.5, 0.19715007, 0.]
    assert_allclose(data[1, :], ref)


def test_hanning_numpy():
    """Test Hanning window against 1D numpy version."""

    size = 101
    cen = (size - 1) // 2
    shape = (size, size)
    win = HanningWindow()
    data = win(shape)
    ref1d = np.hanning(shape[0])
    assert_allclose(data[cen, :], ref1d)


def test_tukey():
    win = TukeyWindow(0.5)
    data = win((5, 5))
    ref = [0., 0.63312767, 1., 0.63312767, 0.]
    assert_allclose(data[1, :], ref)


@pytest.mark.skipif('not HAS_SCIPY')
def test_tukey_scipy():
    """Test Tukey window against 1D scipy version."""

    # scipy.signal.tukey was introduced in Scipy v0.16.0
    from scipy.signal import tukey
    size = 101
    cen = (size - 1) // 2
    shape = (size, size)
    alpha = 0.4
    win = TukeyWindow(alpha=alpha)
    data = win(shape)
    ref1d = tukey(shape[0], alpha=alpha)
    assert_allclose(data[cen, :], ref1d)


def test_cosine_bell():
    win = CosineBellWindow(alpha=0.8)
    data = win((7, 7))
    ref = [0., 0., 0.19715007, 0.5, 0.19715007, 0., 0.]
    assert_allclose(data[2, :], ref)


def test_split_cosine_bell():
    win = SplitCosineBellWindow(alpha=0.8, beta=0.2)
    data = win((5, 5))
    ref = [0., 0.3454915, 1., 0.3454915, 0.]
    assert_allclose(data[2, :], ref)


def test_tophat():
    win = TopHatWindow(beta=0.5)
    data = win((5, 5))
    ref = [0., 1., 1., 1., 0.]
    assert_allclose(data[2, :], ref)


def test_invalid_shape():
    with pytest.raises(ValueError):
        win = HanningWindow()
        win((5,))
